# coding=utf-8
"""Luminance data type."""
from __future__ import division

from .base import DataTypeBase


class Luminance(DataTypeBase):
    """Luminance
    """
    _units = ('cd/m2', 'cd/ft2')
    _si_units = ('cd/m2',)
    _ip_units = ('cd/ft2',)
    _min = 0
    _abbreviation = 'Lv'
    _point_in_time = False

    def _cd_m2_to_cd_ft2(self, value):
        return value / 10.7639

    def _cd_ft2_to_cd_m2(self, value):
        return value * 10.7639

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('cd/m2', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit == 'cd/ft2':
            return values, from_unit
        else:
            return self.to_unit(values, 'cd/ft2', from_unit), 'cd/ft2'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit == 'cd/m2':
            return values, from_unit
        else:
            return self.to_unit(values, 'cd/m2', from_unit), 'cd/m2'


class ZenithLuminance(Luminance):
    _abbreviation = 'ZL'



# coding=utf-8
from __future__ import division

from .legend import Legend, LegendParameters, LegendParametersCategorized

from .datatype.base import DataTypeBase

from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry3d.pointvector import Point3D
from ladybug_geometry.geometry3d.plane import Plane


class GraphicContainer(object):
    """Graphic container used to get legends, title locations, and colors for a graphic.

    Args:
        values: A List or Tuple of numerical values that will be used to
            generate the legend and colors.
        min_point: A Point3D object for the minimum of the bounding box around
            the graphic geometry.
        max_point: A Point3D object for the maximum of the  bounding box around
            the graphic geometry.
        legend_parameters: An Optional LegendParameter object to override default
            parameters of the legend. None indicates that default legend parameters
            will be used. (Default: None).
        data_type: Optional DataType from the ladybug datatype subpackage (ie.
            Temperature()) , which will be used to assign default legend properties.
            If None, the legend associated with this object will contain no units
            unless a unit below is specified. (Default: None).
        unit: Optional text string for the units of the values. (ie. 'C'). If None,
            the default units of the data_type will be used. (Default: None).

    Properties:
        * values
        * min_point
        * max_point
        * legend_parameters
        * data_type
        * unit
        * legend
        * value_colors
        * lower_title_location
        * upper_title_location
    """
    __slots__ = ('_legend', '_min_point', '_max_point', '_data_type', '_unit')

    def __init__(self, values, min_point, max_point,
                 legend_parameters=None, data_type=None, unit=None):
        """Initialize graphic container."""
        # check the input points and legend information
        if isinstance(min_point, Point2D):
            min_point = Point3D(min_point.x, min_point.y)
        if isinstance(max_point, Point2D):
            max_point = Point3D(max_point.x, max_point.y)
        assert isinstance(min_point, Point3D), \
            'min_point should be a ladybug Point3D. Got {}'.format(type(min_point))
        assert isinstance(max_point, Point3D), \
            'max_point should be a ladybug Point3D. Got {}'.format(type(max_point))
        self._legend = Legend(values, legend_parameters)
        self._min_point = min_point
        self._max_point = max_point

        # set default legend parameters based on input data_type and unit
        self._data_type = data_type
        self._unit = unit
        if data_type is not None:
            assert isinstance(data_type, DataTypeBase), \
                'data_type should be a ladybug DataType. Got {}'.format(type(data_type))
            if self.legend_parameters.is_title_default:
                unit = unit if unit else data_type.units[0]
                data_type.is_unit_acceptable(unit)
                self.legend_parameters.title = unit if \
                    self.legend_parameters.vertical \
                    else '{} ({})'.format(data_type.name, unit)
            if data_type.unit_descr is not None and \
                    self.legend_parameters.ordinal_dictionary is None and not \
                    isinstance(self.legend_parameters, LegendParametersCategorized):
                self.legend_parameters.ordinal_dictionary = data_type.unit_descr
                sorted_keys = sorted(data_type.unit_descr.keys())
                if self.legend.is_min_default:
                    self.legend_parameters._min = sorted_keys[0]
                if self.legend.is_max_default:
                    self.legend_parameters._max = sorted_keys[-1]
                assert self.legend_parameters._min <= self.legend_parameters._max, \
                    'Legend min is greater than legend max. {} > {}.'.format(
                        self.legend_parameters._min, self.legend_parameters._max)
                if self.legend_parameters.is_segment_count_default:
                    try:  # try to set the number of segments to align with ordinal text
                        min_i = sorted_keys.index(self.legend_parameters.min)
                        max_i = sorted_keys.index(self.legend_parameters.max)
                        self.legend_parameters.segment_count = \
                            len(sorted_keys[min_i:max_i + 1])
                    except IndexError:
                        pass
        elif unit and self.legend_parameters.is_title_default:
            assert isinstance(unit, str), \
                'Expected string for unit. Got {}.'.format(type(unit))
            self.legend_parameters.title = unit

        # set the default segment_height
        if self.legend_parameters.is_segment_height_default:
            s_count = self.legend_parameters.segment_count
            denom = s_count if s_count >= 8 else 8
            if self.legend_parameters.vertical:
                seg_height = float((self._max_point.y - self._min_point.y) / denom)
                if seg_height == 0:
                    seg_height = float((self._max_point.x - self._min_point.x) / denom)
            else:
                seg_height = float((self._max_point.x - self._min_point.x) / (denom * 2))
                if seg_height == 0:
                    seg_height = float((self._max_point.y - self._min_point.y) / denom)
            self.legend_parameters.properties_3d.segment_height = seg_height
            self.legend_parameters.properties_3d._is_segment_height_default = True

        # set the default segment_width
        if self.legend_parameters.is_segment_width_default:
            if self.legend_parameters.vertical:
                seg_width = self.legend_parameters.segment_height / 2
            else:
                seg_width = self.legend_parameters.text_height * \
                    (len(str(int(self.legend_parameters.max))) +
                     self.legend_parameters.decimal_count + 2)
            self.legend_parameters.properties_3d.segment_width = seg_width
            self.legend_parameters.properties_3d._is_segment_width_default = True

        # set the default base point
        if self.legend_parameters.is_base_plane_default:
            if self.legend_parameters.vertical:
                base_pt = Point3D(
                    self._max_point.x + self.legend_parameters.segment_width,
                    self._min_point.y, self._min_point.z)
            else:
                base_pt = Point3D(
                    self._max_point.x,
                    self._max_point.y + 3 * self.legend_parameters.text_height,
                    self._min_point.z)
            self.legend_parameters.properties_3d.base_plane = Plane(o=base_pt)
            self.legend_parameters.properties_3d._is_base_plane_default = True

    @classmethod
    def from_dict(cls, data):
        """Create a graphic container from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "values": [0, 10],
            "min_point": {"x": 0, "y": 0, "z": 0},
            "max_point": {"x": 10, "y": 10, "z": 0},
            "legend_parameters": {},  # optional LegendParameter specification
            "data_type": {},  # optional DataType object
            "unit": "C"  # optional text for the units
            }
        """
        legend_parameters = None
        if 'legend_parameters' in data and data['legend_parameters'] is not None:
            if data['legend_parameters']['type'] == 'LegendParametersCategorized':
                legend_parameters = LegendParametersCategorized.from_dict(
                    data['legend_parameters'])
            else:
                legend_parameters = LegendParameters.from_dict(data['legend_parameters'])

        data_type = None
        if 'data_type' in data and data['data_type'] is not None:
            data_type = DataTypeBase.from_dict(data['data_type'])
        unit = data['unit'] if 'unit' in data else None

        return cls(data['values'], Point3D.from_dict(data['min_point']),
                   Point3D.from_dict(data['max_point']),
                   legend_parameters, data_type, unit)

    @property
    def values(self):
        """The assigned data set of values."""
        return self._legend.values

    @property
    def min_point(self):
        """Point3D for the minimum of the bounding box around referenced geometry."""
        return self._min_point

    @property
    def max_point(self):
        """Point3D for the maximum of the bounding box around referenced geometry."""
        return self._max_point

    @property
    def legend_parameters(self):
        """The legend parameters assigned to this graphic."""
        return self._legend._legend_par

    @property
    def data_type(self):
        """The data_type input to this object (if it exists)."""
        return self._data_type

    @property
    def unit(self):
        """The unit input to this object (if it exists)."""
        return self._unit

    @property
    def legend(self):
        """The legend assigned to this graphic."""
        return self._legend

    @property
    def value_colors(self):
        """A List of colors associated with the assigned values."""
        return self._legend.value_colors

    @property
    def lower_title_location(self):
        """A Plane for the lower location of title text."""
        return Plane(o=Point3D(
            self._min_point.x,
            self._min_point.y - 2.5 * self._legend.legend_parameters.text_height,
            self._min_point.z))

    @property
    def upper_title_location(self):
        """A Plane for the upper location of title text."""
        return Plane(o=Point3D(
            self._min_point.x,
            self._max_point.y + self._legend.legend_parameters.text_height,
            self._min_point.z))

    def to_dict(self):
        """Get graphic container as a dictionary."""
        base = {
            'type': 'GraphicContainer',
            'values': self.values,
            'min_point': self.min_point.to_dict(),
            'max_point': self.max_point.to_dict(),
            'legend_parameters': self.legend_parameters.to_dict()
        }
        if self.data_type is not None:
            base['data_type'] = self.data_type.to_dict()
        if self.unit is not None:
            base['unit'] = self.unit
        return base

    def __len__(self):
        """Return length of values on the object."""
        return len(self._legend._values)

    def __getitem__(self, key):
        """Return one of the values."""
        return self._legend._values[key]

    def __iter__(self):
        """Iterate through the values."""
        return iter(self._legend._values)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """GraphicContainer representation."""
        return 'Graphic Container ({} values)'.format(len(self))



# coding=utf-8
"""Utility functions for running the functions of CLI commands outside of the CLI."""
import sys
import os


def run_command_function(command_function, arguments=None, options=None):
    """Run a function used within the Ladybug Tools CLI.

    Args:
        command_function: The CLI function to be executed.
        arguments: A list of required arguments for the command_function. None can
            be used if the command function has no required arguments. (Default: None).
        options: A dictionary of options to be passed to the function. The keys
            of this dictionary should be the full names of the options or flags
            in the CLI and formatted with dashes (eg. --output-file). The values
            of the dictionary should be the values to be passed for each of
            the options. For the case of flag arguments, the value should simply
            be an empty string. None can be used here to indicate that all default
            values for options should be used. (Default: None).
    """
    # process the arguments and options
    args = () if arguments is None else arguments
    kwargs = {}
    if options is not None:
        for key, val in options.items():
            clean_key = key[2:] if key.startswith('--') else key
            clean_key = clean_key.replace('-', '_')
            clean_val = True if val == '' else val
            kwargs[clean_key] = clean_val

    # run the command using arguments and keyword arguments
    return command_function(*args, **kwargs)


def process_content_to_output(content_str, output_file=None):
    """Process output strings from commands for various formats of output_files.

    Args:
        content_str: A text string for file contents that are being output from
            a command.
        output_file: Any of the typically supported --output-file types of the
            CLI. This can be a string for a file path, a file object, or the stdout
            file object used by click. If None, the string is simply returned from
            this method. (Default: None).
    """
    if output_file is None:
        return content_str
    elif isinstance(output_file, str):
        dir_name = os.path.dirname(os.path.abspath(output_file))
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name)
        if (sys.version_info < (3, 0)):  # we need to manually encode it as UTF-8
            with open(output_file, 'w') as of:
                of.write(content_str.encode('utf-8'))
        else:
            with open(output_file, 'w', encoding='utf-8') as of:
                of.write(content_str)
    else:
        if 'stdout' not in str(output_file):
            dir_name = os.path.dirname(os.path.abspath(output_file.name))
            if not os.path.isdir(dir_name):
                os.makedirs(dir_name)
        output_file.write(content_str)



# coding=utf-8
"""Class to subdivide the sphere and hemisphere for view-based and radiation studies."""
from __future__ import division

import math

from ladybug_geometry.geometry2d.pointvector import Vector2D
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.mesh import Mesh3D


class ViewSphere(object):
    """Class for subdividing the sphere and hemisphere for view-based studies.

    Note:
        [1] Tregenza, Peter. (1987). Subdivision of the sky hemisphere for luminance
        measurements. Lighting Research & Technology - LIGHTING RES TECHNOL.
        19. 13-14. 10.1177/096032718701900103.

    Properties:
        * tregenza_dome_vectors
        * tregenza_sphere_vectors
        * tregenza_dome_mesh
        * tregenza_dome_mesh_high_res
        * tregenza_sphere_mesh
        * tregenza_solid_angles
        * reinhart_dome_vectors
        * reinhart_sphere_vectors
        * reinhart_dome_mesh
        * reinhart_sphere_mesh
        * reinhart_solid_angles
    """
    # number of patches in each row of the most-used sky domes
    TREGENZA_PATCHES_PER_ROW = (30, 30, 24, 24, 18, 12, 6)
    REINHART_PATCHES_PER_ROW = (60, 60, 60, 60, 48, 48, 48, 48, 36, 36, 24, 24, 12, 12)
    # number of steradians for the patches of each row of the most-used sky domes
    TREGENZA_COEFFICIENTS = \
        (0.0435449227, 0.0416418006, 0.0473984151, 0.0406730411, 0.0428934136,
         0.0445221864, 0.0455168385, 0.0344199465)
    REINHART_COEFFICIENTS = \
        (0.0113221971, 0.0111894547, 0.0109255262, 0.0105335058, 0.0125224872,
         0.0117312774, 0.0108025291, 0.00974713106, 0.011436609, 0.00974295956,
         0.0119026242, 0.00905126163, 0.0121875626, 0.00612971396, 0.00921483254)

    __slots__ = ('_tregenza_dome_vectors', '_tregenza_sphere_vectors',
                 '_tregenza_dome_mesh', '_tregenza_dome_mesh_high_res',
                 '_tregenza_sphere_mesh', '_tregenza_solid_angles',
                 '_reinhart_dome_vectors', '_reinhart_sphere_vectors',
                 '_reinhart_dome_mesh', '_reinhart_sphere_mesh',
                 '_reinhart_solid_angles')

    def __init__(self):
        """Create the ViewSphere."""
        # everything starts with None and properties are generated as requested
        self._tregenza_dome_vectors = None
        self._tregenza_sphere_vectors = None
        self._tregenza_dome_mesh = None
        self._tregenza_dome_mesh_high_res = None
        self._tregenza_sphere_mesh = None
        self._tregenza_solid_angles = None
        self._reinhart_dome_vectors = None
        self._reinhart_sphere_vectors = None
        self._reinhart_dome_mesh = None
        self._reinhart_sphere_mesh = None
        self._reinhart_solid_angles = None

    @property
    def tregenza_dome_vectors(self):
        """An array of 145 vectors representing the Tregenza sky dome."""
        if self._tregenza_dome_vectors is None:
            self._tregenza_dome_mesh, self._tregenza_dome_vectors = self.dome_patches()
        return self._tregenza_dome_vectors

    @property
    def tregenza_sphere_vectors(self):
        """An array of 290 vectors representing a sphere of Tregenza vectors."""
        if self._tregenza_sphere_vectors is None:
            self._tregenza_sphere_mesh, self._tregenza_sphere_vectors = \
                self.sphere_patches()
        return self._tregenza_sphere_vectors

    @property
    def tregenza_dome_mesh(self):
        """An Mesh3D representing the Tregenza sky dome.

        There is one quad face per patch except for the last circular patch, which
        is represented by 6 triangles.
        """
        if self._tregenza_dome_mesh is None:
            self._tregenza_dome_mesh, self._tregenza_dome_vectors = self.dome_patches()
        return self._tregenza_dome_mesh

    @property
    def tregenza_dome_mesh_high_res(self):
        """An high-resolution Mesh3D representing the Tregenza sky dome.

        Each patch is represented by a 3x3 set of quad faces except for the last
        circular patch, which is represented by 18 triangles.
        """
        if self._tregenza_dome_mesh_high_res is None:
            self._tregenza_dome_mesh_high_res, _ = self.dome_patches(3, True)
        return self._tregenza_dome_mesh_high_res

    @property
    def tregenza_sphere_mesh(self):
        """An Mesh3D representing a Tregenza sphere.

        There is one quad face per patch except for the two circular patches, which
        are each represented by 6 triangles.
        """
        if self._tregenza_sphere_mesh is None:
            self._tregenza_sphere_mesh, self._tregenza_sphere_vectors = \
                self.sphere_patches()
        return self._tregenza_sphere_mesh

    @property
    def tregenza_solid_angles(self):
        """Get a list of solid angles that align with the tregenza_dome_vectors."""
        if self._reinhart_solid_angles is None:
            angles = view_sphere.TREGENZA_COEFFICIENTS
            patch_rows = view_sphere.TREGENZA_PATCHES_PER_ROW + (1,)
            patch_angles = []
            for ang, p_count in zip(angles, patch_rows):
                patch_angles.extend([ang] * p_count)
            self._reinhart_solid_angles = tuple(patch_angles)
        return self._reinhart_solid_angles

    @property
    def reinhart_dome_vectors(self):
        """An array of 577 vectors representing the Reinhart sky dome."""
        if self._reinhart_dome_vectors is None:
            self._reinhart_dome_mesh, self._reinhart_dome_vectors = self.dome_patches(2)
        return self._reinhart_dome_vectors

    @property
    def reinhart_sphere_vectors(self):
        """An array of 1154 vectors representing a sphere of Reinhart vectors."""
        if self._reinhart_sphere_vectors is None:
            self._reinhart_sphere_mesh, self._reinhart_sphere_vectors = \
                self.sphere_patches(2)
        return self._reinhart_sphere_vectors

    @property
    def reinhart_dome_mesh(self):
        """An Mesh3D representing the Reinhart sky dome.

        There is one quad face per patch except for the last circular patch, which
        is represented by 12 triangles.
        """
        if self._reinhart_dome_mesh is None:
            self._reinhart_dome_mesh, self._reinhart_dome_vectors = self.dome_patches(2)
        return self._reinhart_dome_mesh

    @property
    def reinhart_sphere_mesh(self):
        """An Mesh3D representing a Reinhart sphere.

        There is one quad face per patch except for the two circular patches, which
        are each represented by 12 triangles.
        """
        if self._reinhart_sphere_mesh is None:
            self._reinhart_sphere_mesh, self._reinhart_sphere_vectors = \
                self.sphere_patches(2)
        return self._reinhart_sphere_mesh

    @property
    def reinhart_solid_angles(self):
        """Get a list of solid angles that align with the reinhart_dome_vectors."""
        if self._reinhart_solid_angles is None:
            angles = view_sphere.REINHART_COEFFICIENTS
            patch_rows = view_sphere.REINHART_PATCHES_PER_ROW + (1,)
            patch_angles = []
            for ang, p_count in zip(angles, patch_rows):
                patch_angles.extend([ang] * p_count)
            self._reinhart_solid_angles = tuple(patch_angles)
        return self._reinhart_solid_angles

    def horizontal_radial_vectors(self, vector_count):
        """Get perfectly horizontal Vector3Ds radiating outward in a circle.

        Args:
            vector_count: An integer for the number of vectors to generate in the
                horizontal plane. This can align with any of the dome or sphere
                patches by setting this to 30 * division_count.

        Returns:
            A list of ladybug_geometry horizontal Vector3D radiating outward in
            a circle. All vectors are unit vectors.
        """
        base_vec = Vector3D(0, 1, 0)
        horiz_angle = -2 * math.pi / vector_count
        return tuple(base_vec.rotate_xy(horiz_angle * i) for i in range(vector_count))

    def horizontal_radial_patches(self, offset_angle=30, division_count=1,
                                  subdivide_in_place=False):
        """Get Vector3Ds within a certain angle offset from the horizontal plane.

        Args:
            offset_angle: A number between 0 and 90 for the angle offset from the
                horizontal plane at which vectors will be included. Vectors both
                above and below this angle will be included (Default: 30 for the
                rough vertical limit of human peripheral vision).
            division_count: A positive integer for the number of times that the
                original Tregenza patches are subdivided. 1 indicates that the
                original Tregenza patches will be used, 2 indicates
                the Reinhart patches will be used, and so on. (Default: 1).
            subdivide_in_place: A boolean to note whether patches should be
                subdivided according to the extension of Tregenza's original
                logic through the Reinhart method (False) or they should be
                simply divided into 4 in place (True).

        Returns:
            A tuple with two elements

            -   patch_mesh: A ladybug_geometry Mesh3D that represents the patches at
                the input division_count. There is one quad face per patch.

            -   patch_vectors: A list of ladybug_geometry Vector3D with one vector
                per patch. These will align with the faces of the patch_mesh.
                All vectors are unit vectors.
        """
        # figure out how many rows and patches should be in the output
        patch_row_count = self._patch_row_count_array(division_count)
        patch_count = self._patch_count_in_radial_offset(
            offset_angle, division_count, patch_row_count, subdivide_in_place)

        # get the dome and vectors and remove faces up tot the patch count
        m_all, v_all = self.dome_patches(division_count, subdivide_in_place)
        pattern = [True] * patch_count + \
            [False] * (sum(patch_row_count) - patch_count + 6 * division_count)
        m_top, _ = m_all.remove_faces(pattern)
        v_top = tuple(vec for vec, val in zip(v_all, pattern) if val)

        # reverse the vectors and negate all the z values of the sky patch mesh
        return self._generate_bottom_from_top(m_top, v_top)

    def horizontal_radial_patch_weights(self, offset_angle=30, division_count=1):
        """Get a list of numbers corresponding to the area weight of each radial patch.

        Args:
            offset_angle: A number between 0 and 90 for the angle offset from the
                horizontal plane at which vectors will be included. Vectors both
                above and below this angle will be included (Default: 30).
            division_count: A positive integer for the number of times that the
                original Tregenza patches are subdivided. (Default: 1).

        Returns:
            A list of numbers with a value for each patch that corresponds to the
            area of that patch. The average value of all the patches is equal to 1.
        """
        # get the areas of the patches and the number of patches to include in the offset
        patch_areas, patch_row_count = self._dome_patch_areas(division_count)
        patch_count = self._patch_count_in_radial_offset(
            offset_angle, division_count, patch_row_count)
        # normalize the patch areas so that they average to 1
        relevant_patches = patch_areas[:patch_count]
        avg_patch_area = sum(relevant_patches) / len(relevant_patches)
        return [p_area / avg_patch_area for p_area in relevant_patches] * 2

    def dome_patches(self, division_count=1, subdivide_in_place=False):
        """Get Vector3Ds and a corresponding Mesh3D for a dome.

        Args:
            division_count: A positive integer for the number of times that the
                original Tregenza patches are subdivided. 1 indicates that the
                original Tregenza patches will be used, 2 indicates
                the Reinhart patches will be used, and so on. (Default: 1).
            subdivide_in_place: A boolean to note whether patches should be
                subdivided according to the extension of Tregenza's original
                logic through the Reinhart method (False) or they should be
                simply divided into 4 in place (True). The latter is useful
                for making higher resolution Mesh visualizations of an
                inherently low-resolution dome.

        Returns:
            A tuple with two elements

            -   patch_mesh: A ladybug_geometry Mesh3D that represents the dome at
                the input division_count. There is one quad face per patch except
                for the last circular patch, which is represented by a number of
                triangles equal to division_count * 6.

            -   patch_vectors: A list of ladybug_geometry Vector3D with one vector
                per patch. These will align with the faces of the patch_mesh up
                until the last circular patch, which will have a single vector
                for the several triangular faces. All vectors are unit vectors.
        """
        # compute constants to be used in the generation of patch points
        patch_row_count = self._patch_row_count_array(division_count)
        base_vec = Vector3D(0, 1, 0)
        rotate_axis = Vector3D(1, 0, 0)
        vertical_angle = math.pi / (2 * len(patch_row_count) + division_count) if \
            subdivide_in_place else math.pi / (2 * len(patch_row_count) + 1)

        # loop through the patch values and generate points for each vertex
        vertices, faces = [], []
        pt_i = -2  # track the number of vertices in the mesh
        for row_i, row_count in enumerate(patch_row_count):
            pt_i += 2  # advance the number of vertices by two
            horiz_angle = -2 * math.pi / row_count  # horizontal angle of each patch
            vec01 = base_vec.rotate(rotate_axis, vertical_angle * row_i)
            vec02 = vec01.rotate(rotate_axis, vertical_angle)
            correction_angle = -horiz_angle / 2
            if subdivide_in_place:
                correction_angle * division_count
            vec1 = vec01.rotate_xy(correction_angle)
            vec2 = vec02.rotate_xy(correction_angle)
            vertices.extend((Point3D(v.x, v.y, v.z) for v in (vec1, vec2)))
            for _ in range(row_count):  # generate the row of patches
                vec3 = vec1.rotate_xy(horiz_angle)
                vec4 = vec2.rotate_xy(horiz_angle)
                vertices.extend((Point3D(v.x, v.y, v.z) for v in (vec3, vec4)))
                faces.append((pt_i, pt_i + 1, pt_i + 3, pt_i + 2))
                pt_i += 2  # advance the number of vertices by two
                vec1, vec2 = vec3, vec4  # reset vec1 and vec2 for the next patch

        # add triangular faces to represent the last circular patch
        end_vert_i = len(vertices)
        start_vert_i = len(vertices) - patch_row_count[-1] * 2 - 1
        vertices.append(Point3D(0, 0, 1))
        for tr_i in range(0, patch_row_count[-1] * 2, 2):
            faces.append((start_vert_i + tr_i, end_vert_i, start_vert_i + tr_i + 2))

        # create the Mesh3D object and derive the patch vectors from the mesh
        patch_mesh = Mesh3D(vertices, faces)
        patch_vectors = patch_mesh.face_normals[:-patch_row_count[-1]] + \
            (Vector3D(0, 0, 1),)
        return patch_mesh, patch_vectors

    def dome_patch_weights(self, division_count=1):
        """Get a list of numbers corresponding to the area weight of each dome patch.

        Args:
            division_count: A positive integer for the number of times that the
                original Tregenza patches are subdivided. (Default: 1).

        Returns:
            A list of numbers with a value for each patch that corresponds to the
            area of that patch. The average value of all the patches is equal to 1.
        """
        # get the areas of the patches
        patch_areas, _ = self._dome_patch_areas(division_count)
        # normalize the patch areas so that they average to 1
        avg_patch_area = 2 * math.pi / len(patch_areas)
        return [p_area / avg_patch_area for p_area in patch_areas]

    def sphere_patches(self, division_count=1, subdivide_in_place=False):
        """Get Vector3Ds and a corresponding Mesh3D for a sphere.

        Args:
            division_count: A positive integer for the number of times that the
                original Tregenza patches are subdivided. 1 indicates that the
                original Tregenza patches will be used, 2 indicates
                the Reinhart patches will be used, and so on. (Default: 1).
            subdivide_in_place: A boolean to note whether patches should be
                subdivided according to the extension of Tregenza's original
                logic through the Reinhart method (False) or they should be
                simply divided into 4 in place (True). The latter is useful
                for making higher resolution Mesh visualizations of an
                inherently low-resolution dome.

        Returns:
            A tuple with two elements

            -   patch_mesh: A ladybug_geometry Mesh3D that represents the sphere at
                the input division_count. There is one quad face per patch except
                for the last circular patch of each hemisphere, which is represented
                by a number of triangles equal to division_count * 6.

            -   patch_vectors: A list of ladybug_geometry Vector3D with one vector
                per patch. These will align with the faces of the patch_mesh except
                for the two circular patches, which will have a single vector
                for the several triangular faces. All vectors are unit vectors.
        """
        # generate patches for the hemisphere
        m_top, v_top = self.dome_patches(division_count, subdivide_in_place)
        # reverse the vectors and negate all the z values of the sky patch mesh
        return self._generate_bottom_from_top(m_top, v_top)

    def sphere_patch_weights(self, division_count=1):
        """Get a list of numbers corresponding to the area weight of each sphere patch.

        Args:
            division_count: A positive integer for the number of times that the
                original Tregenza patches are subdivided. (Default: 1).

        Returns:
            A list of numbers with a value for each patch that corresponds to the
            area of that patch. The average value of all the patches is equal to 1.
        """
        # get the areas of the patches
        patch_areas, _ = self._dome_patch_areas(division_count)
        # normalize the patch areas so that they average to 1
        avg_patch_area = 2 * math.pi / len(patch_areas)
        return [p_area / avg_patch_area for p_area in patch_areas] * 2

    def dome_radial_patches(self, azimuth_count=72, altitude_count=18):
        """Get Vector3Ds and a corresponding Mesh3D for a a radial dome.

        Args:
            azimuth_count: A positive integer for the number of times that
                the horizontal circle will be subdivided into azimuth
                patches. (Default: 72).
            altitude_count: A positive integer for the number of times that
                the dome quarter-circle will be subdivided into altitude
                patches. (Default: 18).

        Returns:
            A tuple with two elements

            -   patch_mesh: A ladybug_geometry Mesh3D that represents the patches at
                the input azimuth_count and altitude_count.

            -   patch_vectors: A list of ladybug_geometry Vector3D with one vector
                per mesh face. These will align with the faces of the patch_mesh.
                All vectors are unit vectors.
        """
        # set up starting vectors and points
        base_vec, rotate_axis = Vector3D(0, 1, 0), Vector3D(1, 0, 0)
        horiz_angle = -2 * math.pi / azimuth_count
        vertical_angle = math.pi / (2 * altitude_count)

        # loop through the patch values and generate points for each vertex
        vertices, faces = [], []
        pt_i = -2  # track the number of vertices in the mesh
        for row_i in range(altitude_count - 1):
            pt_i += 2  # advance the number of vertices by two
            vec1 = base_vec.rotate(rotate_axis, vertical_angle * row_i)
            vec2 = vec1.rotate(rotate_axis, vertical_angle)
            vertices.extend((Point3D(v.x, v.y, v.z) for v in (vec1, vec2)))
            for _ in range(azimuth_count):  # generate the row of patches
                vec3 = vec1.rotate_xy(horiz_angle)
                vec4 = vec2.rotate_xy(horiz_angle)
                vertices.extend((Point3D(v.x, v.y, v.z) for v in (vec3, vec4)))
                faces.append((pt_i, pt_i + 1, pt_i + 3, pt_i + 2))
                pt_i += 2  # advance the number of vertices by two
                vec1, vec2 = vec3, vec4  # reset vec1 and vec2 for the next patch

        # add triangular faces to represent the last circular patch
        end_vert_i = len(vertices)
        start_vert_i = len(vertices) - azimuth_count * 2 - 1
        vertices.append(Point3D(0, 0, 1))
        for tr_i in range(0, azimuth_count * 2, 2):
            faces.append((start_vert_i + tr_i, end_vert_i, start_vert_i + tr_i + 2))

        # create the Mesh3D object and derive the patch vectors from the mesh
        patch_mesh = Mesh3D(vertices, faces)
        patch_vectors = patch_mesh.face_normals
        return patch_mesh, patch_vectors

    def dome_radial_patch_weights(self, azimuth_count=72, altitude_count=18):
        """Get a list of numbers corresponding to the area weight of each dome patch.

        Args:
            azimuth_count: A positive integer for the number of times that
                the horizontal circle will be subdivided into azimuth
                patches. (Default: 72).
            altitude_count: A positive integer for the number of times that
                the dome quarter-circle will be subdivided into altitude
                patches. (Default: 18).

        Returns:
            A list of numbers with a value for each patch that corresponds to the
            area of that patch. The average value of all the patches is equal to 1.
        """
        # get the areas of the patches
        patch_areas = self._dome_radial_patch_areas(azimuth_count, altitude_count)
        # normalize the patch areas so that they average to 1
        total_patch_area = 2 * math.pi
        return [p_area / total_patch_area for p_area in patch_areas]

    def horizontal_circle_view_mesh(
            self, center_point=Point3D(0, 0, 0), radius=1, azimuth_count=72):
        """Get a mesh of a horizontal circle with vertices coordinated with view_vectors.

        Args:
            center_point: A Point3D for the center of the mesh. (Default: (0, 0, 0)).
            radius: A number for the radius of the circle. (Default: 1).
            azimuth_count: A positive integer greater than or equal to 3 for
                the number of times that the horizontal circle will be
                subdivided into vertices. (Default: 72).

        Returns:
            A tuple with two elements

            -   circle_mesh: A ladybug_geometry circular Mesh3D that represents
                the horizontal view at the input azimuth_count.

            -   view_vecs: A tuple of ladybug_geometry Vector3D with one vector
                per mesh vertex. The first vertex of the circle_mesh is the center
                and each one after that is coordinated with the vector here.
        """
        # generate a list of vectors over the circle
        view_vecs = self.horizontal_radial_vectors(azimuth_count)
        # use the direction vectors to create a mesh of the sky dome
        vertices, faces = [center_point], []
        for i, vec in enumerate(view_vecs):
            vertices.append(center_point.move(vec * radius))
            faces.append((0, i + 1, i + 2))
        faces.pop(-1)
        faces.append((0, azimuth_count, 1))
        circle_mesh = Mesh3D(vertices, faces)
        return circle_mesh, view_vecs

    def horizontal_radial_view_mesh(
            self, center_point=Point3D(0, 0, 0), radius=1, offset_angle=30,
            azimuth_count=72, altitude_count=6):
        """Get a mesh of a radial circle with vertices coordinated with view_vectors.

        Args:
            center_point: A Point3D for the center of the mesh. (Default: (0, 0, 0)).
            radius: A number for the radius of the circle. (Default: 1).
            offset_angle: A number between 0 and 90 for the angle offset from the
                horizontal plane at which vectors will be included. Vectors both
                above and below this angle will be included (Default: 30).
            azimuth_count: A positive integer greater than or equal to 3 for the
                number of times that the horizontal circle will be subdivided
                into vertices. (Default: 72).
            altitude_count: An integer greater than or equal to 1, which notes
                the number of vertical orientations at which the altitude will
                be evaluated. (Default: 18).

        Returns:
            A tuple with two elements

            -   radial_mesh: A ladybug_geometry Mesh3D that represents the horizontal
                radial view at the input azimuth_count and altitude_count.

            -   view_vecs: A list of ladybug_geometry Vector3D with one vector
                per mesh vertex. The first vertex of the radial_mesh is the center
                and each one after that is coordinated with the vector here.
        """
        # compute the global parameters for generating the mesh
        horiz_angle = -2 * math.pi / azimuth_count
        vert_angle = (math.radians(offset_angle)) / altitude_count
        base_vec, x_axis = Vector3D(0, 1, 0), Vector3D(1, 0, 0)

        # generate a list of vectors over the circle
        view_vecs = list(self.horizontal_radial_vectors(azimuth_count))
        vertices, faces = [center_point], []
        for vec in view_vecs:
            vertices.append(center_point.move(vec * radius))

        # generate a list of vectors over the horizontal radial domain
        up_i1, up_i2, up_i3, up_i4 = 2, 1, azimuth_count + 1, azimuth_count + 3
        dn_i1, dn_i2, dn_i3, dn_i4 = 1, 2, azimuth_count + 4, azimuth_count + 2
        for v in range(1, altitude_count + 1):
            up_vec = base_vec.rotate(x_axis, vert_angle * v)
            dn_vec = base_vec.rotate(x_axis, vert_angle * -v)
            view_vecs.append(up_vec)
            view_vecs.append(dn_vec)
            vertices.append(center_point.move(up_vec * radius))
            vertices.append(center_point.move(dn_vec * radius))
            vi = 1 if v == 1 else 2
            for h in range(1, azimuth_count):
                vv_up = up_vec.rotate_xy(horiz_angle * h)
                vv_dn = dn_vec.rotate_xy(horiz_angle * h)
                view_vecs.append(vv_up)
                view_vecs.append(vv_dn)
                vertices.append(center_point.move(vv_up * radius))
                vertices.append(center_point.move(vv_dn * radius))
                faces.append((up_i1, up_i2, up_i3, up_i4))
                faces.append((dn_i1, dn_i2, dn_i3, dn_i4))
                up_i1, up_i2, up_i3, up_i4 = up_i1 + vi, up_i2 + vi, up_i3 + 2, up_i4 + 2
                dn_i1, dn_i2, dn_i3, dn_i4 = dn_i1 + vi, dn_i2 + vi, dn_i3 + 2, dn_i4 + 2
            sub_i = azimuth_count if v == 1 else 2 * azimuth_count
            faces.append((up_i1 - sub_i, up_i2, up_i3, up_i4 - (2 * azimuth_count)))
            faces.append((dn_i1, dn_i2 - sub_i, dn_i3 - (2 * azimuth_count), dn_i4))
            ri = 3 if v == 1 else 2
            up_i1, up_i2 = up_i1 + 2, up_i2 + vi
            up_i3, up_i4 = up_i3 + 2, up_i4 + 2
            dn_i1, dn_i2 = dn_i1 + 2, dn_i2 + ri
            dn_i3, dn_i4 = dn_i3 + 2, dn_i4 + 2

        # add a series of triangular faces to fill in the top and bottom of the mesh
        az_2 = (2 * azimuth_count)
        up_i3, up_i4 = up_i3 - az_2, up_i4 - az_2
        dn_i3, dn_i4 = dn_i3 - az_2, dn_i4 - az_2
        for h in range(azimuth_count - 1):
            faces.append((0, up_i4, up_i3))
            faces.append((0, dn_i4, dn_i3))
            up_i3, up_i4 = up_i3 + 2, up_i4 + 2
            dn_i3, dn_i4 = dn_i3 + 2, dn_i4 + 2
        faces.append((0, up_i4 - az_2, up_i3))
        faces.append((0, dn_i3 - 2, dn_i4 - az_2 + 2))

        radial_mesh = Mesh3D(vertices, faces)
        return radial_mesh, view_vecs

    def dome_view_mesh(
            self, center_point=Point3D(0, 0, 0), radius=1,
            azimuth_count=72, altitude_count=18):
        """Get a mesh of a horizontal circle with vertices coordinated with view_vectors.

        Args:
            center_point: A Point3D for the center of the mesh. (Default: (0, 0, 0)).
            radius: A number for the radius of the circle. (Default: 1).
            azimuth_count: An integer greater than or equal to 3, which notes the number
                of horizontal orientations to be evaluated on the dome. (Default: 72).
            altitude_count: An integer greater than or equal to 3, which notes the number
                of vertical orientations to be evaluated on the dome. (Default: 18).

        Returns:
            A tuple with two elements

            -   dome_mesh: A ladybug_geometry Mesh3D that represents the hemispherical
                view dome at the input azimuth_count and altitude_count.

            -   view_vecs: A list of ladybug_geometry Vector3D with one vector
                per mesh vertex.
        """
        # generate a list of vectors over the dome
        horiz_angle = -2 * math.pi / azimuth_count
        vert_angle = (math.pi / 2) / altitude_count
        view_vecs = []
        for v in range(altitude_count):
            x_axis = Vector3D(1, 0, 0)
            base_vec = Vector3D(0, 1, 0)
            n_vec = base_vec.rotate(x_axis, vert_angle * v)
            for h in range(azimuth_count):
                view_vecs.append(n_vec.rotate_xy(horiz_angle * h))
        view_vecs.append(Vector3D(0, 0, 1))

        # use the direction vectors to create a mesh of the sky dome
        vertices = []
        for vec in view_vecs:
            vertices.append(center_point.move(vec * radius))
        faces, pt_i, az_ct = [], 0, azimuth_count
        for _ in range(altitude_count - 1):
            for _ in range(az_ct - 1):
                faces.append((pt_i, pt_i + 1, pt_i + az_ct + 1, pt_i + az_ct))
                pt_i += 1  # advance the number of vertices
            faces.append((pt_i, pt_i - az_ct + 1, pt_i + 1, pt_i + az_ct))
            pt_i += 1  # advance the number of vertices
        # add triangular faces to represent the last circular patch
        end_vert_i = len(vertices) - 1
        start_vert_i = len(vertices) - azimuth_count - 1
        for tr_i in range(0, azimuth_count - 1):
            faces.append((start_vert_i + tr_i, end_vert_i, start_vert_i + tr_i + 1))
        faces.append((end_vert_i - 1, end_vert_i, start_vert_i))
        dome_mesh = Mesh3D(vertices, faces)
        return dome_mesh, view_vecs

    def sphere_view_mesh(
            self, center_point=Point3D(0, 0, 0), radius=1,
            azimuth_count=72, altitude_count=18):
        """Get a mesh of a sphere with vertices coordinated with view_vectors.

        Args:
            center_point: A Point3D for the center of the mesh. (Default: (0, 0, 0)).
            radius: A number for the radius of the sphere. (Default: 1).
            azimuth_count: An integer greater than or equal to 3, which notes the number
                of horizontal orientations to be evaluated on the sphere. (Default: 72).
            altitude_count: An integer greater than or equal to 3, which notes the number
                of vertical orientations to be evaluated on the sphere. (Default: 18).

        Returns:
            A tuple with two elements

            -   sphere_mesh: A ladybug_geometry Mesh3D that represents the
                view sphere at the input azimuth_count and altitude_count.

            -   view_vecs: A list of ladybug_geometry Vector3D with one vector
                per mesh vertex.
        """
        # compute the global parameters for generating the mesh
        horiz_angle = -2 * math.pi / azimuth_count
        vert_angle = (math.pi / 2) / altitude_count
        base_vec, x_axis = Vector3D(0, 1, 0), Vector3D(1, 0, 0)

        # generate a list of vectors over the circle
        view_vecs = list(self.horizontal_radial_vectors(azimuth_count))
        vertices, faces = [], []
        for vec in view_vecs:
            vertices.append(center_point.move(vec * radius))

        # generate a list of vectors over the horizontal radial domain
        up_i1, up_i2, up_i3, up_i4 = 1, 0, azimuth_count, azimuth_count + 2
        dn_i1, dn_i2, dn_i3, dn_i4 = 0, 1, azimuth_count + 3, azimuth_count + 1
        for v in range(1, altitude_count):
            up_vec = base_vec.rotate(x_axis, vert_angle * v)
            dn_vec = base_vec.rotate(x_axis, vert_angle * -v)
            view_vecs.append(up_vec)
            view_vecs.append(dn_vec)
            vertices.append(center_point.move(up_vec * radius))
            vertices.append(center_point.move(dn_vec * radius))
            vi = 1 if v == 1 else 2
            for h in range(1, azimuth_count):
                vv_up = up_vec.rotate_xy(horiz_angle * h)
                vv_dn = dn_vec.rotate_xy(horiz_angle * h)
                view_vecs.append(vv_up)
                view_vecs.append(vv_dn)
                vertices.append(center_point.move(vv_up * radius))
                vertices.append(center_point.move(vv_dn * radius))
                faces.append((up_i1, up_i2, up_i3, up_i4))
                faces.append((dn_i1, dn_i2, dn_i3, dn_i4))
                up_i1, up_i2, up_i3, up_i4 = up_i1 + vi, up_i2 + vi, up_i3 + 2, up_i4 + 2
                dn_i1, dn_i2, dn_i3, dn_i4 = dn_i1 + vi, dn_i2 + vi, dn_i3 + 2, dn_i4 + 2
            sub_i = azimuth_count if v == 1 else 2 * azimuth_count
            faces.append((up_i1 - sub_i, up_i2, up_i3, up_i4 - (2 * azimuth_count)))
            faces.append((dn_i1, dn_i2 - sub_i, dn_i3 - (2 * azimuth_count), dn_i4))
            ri = 3 if v == 1 else 2
            up_i1, up_i2 = up_i1 + 2, up_i2 + vi
            up_i3, up_i4 = up_i3 + 2, up_i4 + 2
            dn_i1, dn_i2 = dn_i1 + 2, dn_i2 + ri
            dn_i3, dn_i4 = dn_i3 + 2, dn_i4 + 2

        # add a series of triangular faces to fill in the top and bottom of the mesh
        top_vec, bot_vec = Vector3D(0, 0, 1), Vector3D(0, 0, -1)
        top_i = len(vertices)
        bot_i = top_i + 1
        view_vecs.append(top_vec)
        view_vecs.append(bot_vec)
        vertices.append(center_point.move(top_vec * radius))
        vertices.append(center_point.move(bot_vec * radius))
        az_2 = (2 * azimuth_count)
        up_i3, up_i4 = up_i3 - az_2, up_i4 - az_2
        dn_i3, dn_i4 = dn_i3 - az_2, dn_i4 - az_2
        for h in range(azimuth_count - 1):
            faces.append((top_i, up_i4, up_i3))
            faces.append((bot_i, dn_i4, dn_i3))
            up_i3, up_i4 = up_i3 + 2, up_i4 + 2
            dn_i3, dn_i4 = dn_i3 + 2, dn_i4 + 2
        faces.append((top_i, up_i4 - az_2, up_i3))
        faces.append((bot_i, dn_i3 - 2, dn_i4 - az_2 + 2))

        radial_mesh = Mesh3D(vertices, faces)
        return radial_mesh, view_vecs

    @staticmethod
    def orientation_pattern(plane_normal, view_vectors):
        """Get booleans for whether view vectors are blocked by a plane.

        Args:
            plane_normal: A Vector3D for the normal of the plane.
            view_vectors: A list of view vectors which will be evaluated to determine
                if they are blocked by the plane or not.

        Returns:
            A tuple with two values.

            -   mask_pattern -- A list of booleans for whether each of the view
                vectors are blocked by the plane (True) or not (False).

            -   angles -- A list of angles in radians for the angle between the
                plane normal and each view vector.
        """
        mask_pattern, angles, max_angle = [], [], math.pi / 2
        for vec in view_vectors:
            ang = vec.angle(plane_normal)
            angles.append(ang)
            mask_pattern.append(ang > max_angle)
        return mask_pattern, angles

    @staticmethod
    def overhang_pattern(plane_normal, overhang_angle, view_vectors):
        """Get booleans for whether a view vectors are blocked by a overhang.

        Args:
            plane_normal: A Vector3D for the normal of the plane.
            overhang_angle: A number between 0 and 90 for the projection angle
                of an overhang in degrees.
            view_vectors: A list of view vectors which will be evaluated to
                determine if they are blocked by the plane or not.

        Returns:
            A list of booleans for whether each of the view vectors are blocked by
            the overhang (True) or not (False).
        """
        overhang_norm = plane_normal.reverse()
        rotation_axis = overhang_norm.rotate_xy(-math.pi / 2)
        rotation_axis = Vector3D(rotation_axis.x, rotation_axis.y, 0)
        overhang_norm = overhang_norm.rotate(rotation_axis, math.radians(overhang_angle))
        max_angle = math.pi / 2
        return [vec.angle(overhang_norm) < max_angle for vec in view_vectors]

    @staticmethod
    def fin_pattern(plane_normal, left_fin_angle, right_fin_angle, view_vectors):
        """Get booleans for whether a view vectors are blocked by left and right fins.

        Args:
            plane_normal: A Vector3D for the normal of the plane.
            left_fin_angle: A number between 0 and 90 for the projection angle of a
                fin on the left side in degrees.
            right_fin_angle: A number between 0 and 90 for the projection angle of a
                fin on the right side in degrees.
            view_vectors: A list of view vectors which will be evaluated to determine
                if they are blocked by the plane or not.

        Returns:
            A list of booleans for whether each of the view vectors are blocked by
            the fins (True) or not (False).
        """
        # get the min and max angles for the area not blocked by fins
        y_axis, norm_2d = Vector2D(0, 1), Vector2D(plane_normal.x, plane_normal.y)
        srf_angle = math.degrees(norm_2d.angle_clockwise(y_axis))
        angle_min = srf_angle - 90 + right_fin_angle \
            if right_fin_angle else srf_angle - 90
        angle_max = srf_angle + 90 - left_fin_angle \
            if left_fin_angle else srf_angle + 90
        if angle_max > 360:
            angle_max, angle_min = angle_max - 360, angle_min - 360
        if angle_max < 0:
            angle_max, angle_min = angle_max + 360, angle_min + 360

        # evaluate the view_vectors in relation to the min and max angle
        mask_pattern = []
        for vec in view_vectors:
            ang = math.degrees(Vector2D(vec.x, vec.y).angle_clockwise(y_axis))
            is_visible = (ang < angle_max and ang > angle_min) if angle_min > 0 else \
                (ang < angle_max or ang > angle_min + 360)
            mask_pattern.append(not is_visible)
        return mask_pattern

    @staticmethod
    def _dome_radial_patch_areas(azimuth_count=72, altitude_count=18):
        """Get the area of each patch in a radial dome."""
        # get the areas of each spherical cap moving up the unit dome
        vert_angle = math.pi / (2 * altitude_count)
        cap_areas = [2 * math.pi]
        current_angle = vert_angle
        for i in range(altitude_count):
            cap_areas.append(2 * math.pi * (1 - math.sin(current_angle)))
            current_angle += vert_angle

        # get the area of each row and subdivide it by the patch count of the row
        row_areas = [cap_areas[i] - cap_areas[i + 1] for i in range(len(cap_areas) - 1)]
        patch_areas = []
        for row_area in row_areas:
            patch_areas.extend([row_area / azimuth_count] * azimuth_count)
        return patch_areas

    @staticmethod
    def _dome_patch_areas(division_count):
        """Get the area of each patch in a dome from a division_count."""
        # get the areas of each spherical cap moving up the unit dome
        patch_row_count = ViewSphere._patch_row_count_array(division_count)
        vert_angle = math.pi / (2 * len(patch_row_count) + division_count)
        cap_areas = [2 * math.pi]
        current_angle = vert_angle
        for i in range(len(patch_row_count)):
            cap_areas.append(2 * math.pi * (1 - math.sin(current_angle)))
            current_angle += vert_angle

        # get the area of each row and subdivide it by the patch count of the row
        row_areas = [cap_areas[i] - cap_areas[i + 1] for i in range(len(cap_areas) - 1)]
        patch_areas = []
        for row_count, row_area in zip(patch_row_count, row_areas):
            patch_areas.extend([row_area / row_count] * row_count)
        patch_areas.append(cap_areas[-1])
        return patch_areas, patch_row_count

    @staticmethod
    def _patch_count_in_radial_offset(offset_angle, division_count, patch_row_count,
                                      subdivide_in_place=False):
        """Get the number of patches within a radial offset from the horizontal plane."""
        rad_angle = math.radians(offset_angle)
        patch_rows = len(patch_row_count)
        vert_angle = math.pi / (2 * patch_rows + division_count) if subdivide_in_place \
            else math.pi / (2 * patch_rows + 1)
        row_count = int(round(rad_angle / vert_angle))
        return sum(patch_row_count[:row_count])

    @staticmethod
    def _patch_row_count_array(division_count):
        """Get an array of the number of patches in each dome row from division_count."""
        patch_row_count = ViewSphere.TREGENZA_PATCHES_PER_ROW
        if division_count != 1:
            patch_row_count = [init_ct * division_count for init_ct in patch_row_count
                               for i in range(division_count)]
        return patch_row_count

    @staticmethod
    def _generate_bottom_from_top(m_top, v_top):
        """Get a joined mesh and vectors for top and bottom from only top vectors."""
        # reverse the vectors and negate all the z values of the sky patch mesh
        verts = tuple(Point3D(pt.x, pt.y, -pt.z) for pt in m_top.vertices)
        faces = tuple(face[::-1] for face in m_top.faces)
        m_bottom = Mesh3D(verts, faces)
        v_bottom = tuple(Vector3D(v.x, v.y, -v.z) for v in v_top)

        # join everything together
        patch_mesh = Mesh3D.join_meshes([m_top, m_bottom])
        patch_vectors = v_top + v_bottom
        return patch_mesh, patch_vectors

    def __repr__(self):
        """ViewSphere representation."""
        return 'ViewSphere'


# make a single object that can be reused throughout the library
view_sphere = ViewSphere()



# coding=utf-8
"""Module converting between magnetic and true North."""
from __future__ import division

import math
import os


class WorldMagneticModel(object):
    """World Magnetic Model (WMM) that can convert from magnetic to true North.

     Args:
        cof_file: The full path to a .COF file containing the coefficients that
            form the inputs for the World Magnetic Model (WMM). A new set of coefficients
            is published roughly every 5 years as the magnetic poles continue to
            move. If None, coefficients will be derived from the WMM.COF file contained
            within this Python package, which should be for the most recent model.
            If not, the most recent coefficients are available at
            https://www.ncei.noaa.gov/products/world-magnetic-model/wmm-coefficients
    """

    def __init__(self, cof_file=None):
        """Initialize WorldMagneticModel."""
        # use the .COF file in this package if not specified
        if cof_file is None:
            cof_file = os.path.join(os.path.dirname(__file__), 'WMM.COF')

        # parse the coefficients from the file contents
        wmm = []
        with open(cof_file) as wmm_file:
            for line in wmm_file:
                vals = line.strip().split()
                if len(vals) == 3:
                    self.epoch = float(vals[0])
                    self.model = vals[1]
                    self.modeldate = vals[2]
                elif len(vals) == 6:
                    line_dict = {
                        'n': int(float(vals[0])),
                        'm': int(float(vals[1])),
                        'gnm': float(vals[2]),
                        'hnm': float(vals[3]),
                        'dgnm': float(vals[4]),
                        'dhnm': float(vals[5])
                    }
                    wmm.append(line_dict)

        # set the global constants used by the WMM
        z = [0.0] * 15
        self.maxord = 12
        self.tc = []
        for _ in range(14):
            self.tc.append(z[0:13])
        self.sp = z[0:14]
        self.cp = z[0:14]
        self.cp[0] = 1.0
        self.pp = z[0:13]
        self.pp[0] = 1.0
        self.p = []
        for _ in range(14):
            self.p.append(z[0:14])
        self.p[0][0] = 1.0
        self.dp = []
        for _ in range(14):
            self.dp.append(z[0:13])
        self.a = 6378.137
        self.b = 6356.7523142
        self.re = 6371.2
        self.a2 = self.a*self.a
        self.b2 = self.b*self.b
        self.c2 = self.a2-self.b2
        self.a4 = self.a2*self.a2
        self.b4 = self.b2*self.b2
        self.c4 = self.a4 - self.b4

        self.c = []
        self.cd = []
        for _ in range(14):
            self.c.append(z[0:14])
            self.cd.append(z[0:14])

        # adjust C and CD from the file contents
        for wmmnm in wmm:
            m = wmmnm['m']
            n = wmmnm['n']
            gnm = wmmnm['gnm']
            hnm = wmmnm['hnm']
            dgnm = wmmnm['dgnm']
            dhnm = wmmnm['dhnm']
            if (m <= n):
                self.c[m][n] = gnm
                self.cd[m][n] = dgnm
                if (m != 0):
                    self.c[n][m - 1] = hnm
                    self.cd[n][m - 1] = dhnm

        # convert schmidt normalized gauss coefficients to un-normalized
        self.snorm = []
        for _ in range(13):
            self.snorm.append(z[0:13])
        self.snorm[0][0] = 1.0
        self.k = []
        for _ in range(13):
            self.k.append(z[0:13])
        self.k[1][1] = 0.0
        self.fn = [float(i) for i in range(14)]
        self.fm = [float(i) for i in range(13)]
        for n in range(1, self.maxord + 1):
            self.snorm[0][n] = self.snorm[0][n - 1] * (2.0 * n - 1) / n
            j = 2.0
            m = 0
            d_1 = 1
            d_2 = (n - m + d_1) / d_1
            while (d_2 > 0):
                self.k[m][n] = (((n - 1) * (n - 1)) - (m * m)) / \
                    ((2.0 * n - 1) * (2.0 * n - 3.0))
                if (m > 0):
                    flnmj = ((n - m + 1.0) * j) / (n + m)
                    self.snorm[m][n] = self.snorm[m - 1][n] * math.sqrt(flnmj)
                    j = 1.0
                    self.c[n][m - 1] = self.snorm[m][n] * self.c[n][m - 1]
                    self.cd[n][m - 1] = self.snorm[m][n] * self.cd[n][m - 1]
                self.c[m][n] = self.snorm[m][n] * self.c[m][n]
                self.cd[m][n] = self.snorm[m][n] * self.cd[m][n]
                d_2 = d_2 - 1
                m = m + d_1

        # set default lat, lon and altitude, which will be overwritten as used
        self.otime = -1000.0
        self.oalt = -1000.0
        self.olat = -1000.0
        self.olon = -1000.0

    def magnetic_declination(self, latitude=0, longitude=0, elevation=0, year=2025):
        """Compute the magnetic declination using the World Magnetic Model (WMM).

        Magnetic declination is the difference between magnetic North and true North at
        a given location on the globe (expressed in terms of degrees). The function
        here uses the same method that underlies the NOAA Magnetic Declination
        calculator.

        Args:
            latitude: A number between -90 and 90 for the latitude of the location
                in degrees. (Default: 0 for the equator).
            longitude: A number between -180 and 180 for the longitude of the location
                in degrees (Default: 0 for the prime meridian).
            elevation: A number for elevation of the location in meters. (Default: 0).
            year: A number for the year in which the magnetic declination is
                being evaluated. Decimal values are accepted. (Default: 2025).

        Returns:
            A number for the magnetic declination in degrees.
        """
        # set up the properties given the location and year information
        alt = elevation / 1000   # to kilometers
        dt = year - self.epoch
        glat = latitude
        glon = longitude
        rlat = math.radians(glat)
        rlon = math.radians(glon)
        srlon = math.sin(rlon)
        srlat = math.sin(rlat)
        crlon = math.cos(rlon)
        crlat = math.cos(rlat)
        srlat2 = srlat * srlat
        crlat2 = crlat * crlat
        self.sp[1] = srlon
        self.cp[1] = crlon

        # convert from geodetic to spherical coordinates
        if (alt != self.oalt or glat != self.olat):
            q = math.sqrt(self.a2 - self.c2 * srlat2)
            q1 = alt * q
            q2 = ((q1 + self.a2) / (q1 + self.b2)) * ((q1 + self.a2) / (q1 + self.b2))
            ct = srlat / math.sqrt(q2 * crlat2 + srlat2)
            st = math.sqrt(1.0 - (ct * ct))
            r2 = (alt * alt) + 2.0 * q1 + (self.a4 - self.c4 * srlat2) / (q * q)
            r = math.sqrt(r2)
            d = math.sqrt(self.a2 * crlat2 + self.b2 * srlat2)
            ca = (alt + d) / r
            sa = self.c2 * crlat * srlat / (r * d)

        if (glon != self.olon):
            for m in range(2, self.maxord + 1):
                self.sp[m] = self.sp[1] * self.cp[m - 1] + self.cp[1] * self.sp[m - 1]
                self.cp[m] = self.cp[1] * self.cp[m - 1] - self.sp[1] * self.sp[m - 1]

        aor = self.re / r
        ar = aor * aor
        br = bt = bp = bpp = 0.0
        for n in range(1, self.maxord + 1):
            ar = ar * aor
            m = 0
            D4 = (n + m + 1)
            while D4 > 0:
                # compute un-normalized associated polynomials and derivatives
                if (alt != self.oalt or glat != self.olat):
                    if (n == m):
                        self.p[m][n] = st * self.p[m - 1][n - 1]
                        self.dp[m][n] = \
                            st * self.dp[m - 1][n - 1] + ct * self.p[m - 1][n - 1]

                    elif (n == 1 and m == 0):
                        self.p[m][n] = ct * self.p[m][n-1]
                        self.dp[m][n] = ct * self.dp[m][n - 1] - st * self.p[m][n - 1]

                    elif (n > 1 and n != m):
                        if (m > n - 2):
                            self.p[m][n - 2] = 0
                        if (m > n - 2):
                            self.dp[m][n - 2] = 0.0
                        self.p[m][n] = \
                            ct * self.p[m][n - 1] - self.k[m][n] * self.p[m][n - 2]
                        self.dp[m][n] = ct * self.dp[m][n - 1] - \
                            st * self.p[m][n - 1]-self.k[m][n] * self.dp[m][n - 2]

                # time adjust the gauss coefficients
                if (year != self.otime):
                    self.tc[m][n] = self.c[m][n] + dt * self.cd[m][n]
                    if (m != 0):
                        self.tc[n][m - 1] = self.c[n][m - 1] + dt * self.cd[n][m - 1]

                # accumulate terms of the spherical harmonic expansions
                par = ar * self.p[m][n]
                if (m == 0):
                    temp1 = self.tc[m][n] * self.cp[m]
                    temp2 = self.tc[m][n] * self.sp[m]
                else:
                    temp1 = self.tc[m][n] * self.cp[m] + self.tc[n][m - 1] * self.sp[m]
                    temp2 = self.tc[m][n] * self.sp[m] - self.tc[n][m - 1] * self.cp[m]

                bt = bt - ar * temp1 * self.dp[m][n]
                bp = bp + (self.fm[m] * temp2 * par)
                br = br + (self.fn[n] * temp1 * par)

                # special case: north/south geographic poles
                if (st == 0.0 and m == 1):
                    if (n == 1):
                        self.pp[n] = self.pp[n - 1]
                    else:
                        self.pp[n] = ct * self.pp[n - 1] - self.k[m][n] * self.pp[n - 2]
                    parp = ar * self.pp[n]
                    bpp = bpp + (self.fm[m] * temp2 * parp)
                D4 = D4 - 1
                m = m + 1

        if (st == 0.0):
            bp = bpp
        else:
            bp = bp/st

        # rotate magnetic vector components from spherical to geodetic coordinates
        bx = -bt * ca - br * sa
        by = bp
        declination = math.degrees(math.atan2(by, bx))

        # set the location attributes that the model has been aligned to
        self.otime = year
        self.oalt = alt
        self.olat = glat
        self.olon = glon

        return declination

    def magnetic_to_true_north(self, location, magnetic_north=0, year=2025):
        """Compute true North from a magnetic North vector.

        Args:
            location: A Ladybug Location object that will be used to determine the
                magnetic declination.
            magnetic_north: A number between -360 and 360 for the counterclockwise
                difference between the North and the positive Y-axis in degrees.
                90 is West and 270 is East (Default: 0).
            year: A number for the year in which the magnetic declination is
                being evaluated. Decimal values are accepted. (Default: 2025).

        Returns:
            A number between -360 and 360 for the true North angle in degrees.
        """
        declination = self.magnetic_declination(
            location.latitude, location.longitude, location.elevation, year)
        true_north = magnetic_north + declination
        if true_north > 360:
            true_north = true_north - 360
        elif true_north < -360:
            true_north = 360 + true_north
        return true_north

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Return WorldMagneticModel as a string."""
        return 'WorldMagneticModel: {}'.format(self.epoch)



# coding=utf-8
"""Conductivity data type."""
from __future__ import division

from .base import DataTypeBase


class Conductivity(DataTypeBase):
    """Conductivity
    """
    _units = ('W/m-K', 'Btu/h-ft-F', 'cal/s-cm-C')
    _si_units = ('W/m-K',)
    _ip_units = ('Btu/h-ft-F',)
    _min = 0
    _abbreviation = 'k'

    def _W_mK_to_Btu_hftF(self, value):
        return value / 1.7295772056

    def _Btu_hftF_to_W_mK(self, value):
        return value * 1.7295772056

    def _W_mK_to_cal_scmC(self, value):
        return value / 418.4

    def _cal_scmC_to_W_mK(self, value):
        return value * 418.4

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('W/m-K', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit == 'Btu/h-ft-F':
            return values, from_unit
        else:
            return self.to_unit(values, 'Btu/h-ft-F', from_unit), 'Btu/h-ft-F'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit == 'W/m-K':
            return values, from_unit
        else:
            return self.to_unit(values, 'W/m-K', from_unit), 'W/m-K'



# coding=utf-8
"""Module for visualization of hourly data collections."""
from __future__ import division

from .analysisperiod import AnalysisPeriod
from .datacollection import HourlyContinuousCollection, HourlyDiscontinuousCollection
from .graphic import GraphicContainer

from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
from ladybug_geometry.geometry2d.line import LineSegment2D
from ladybug_geometry.geometry2d.polygon import Polygon2D
from ladybug_geometry.geometry2d.polyline import Polyline2D
from ladybug_geometry.geometry2d.mesh import Mesh2D
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.line import LineSegment3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.polyline import Polyline3D
from ladybug_geometry.geometry3d.mesh import Mesh3D


class HourlyPlot(object):
    """Object for visualization of hourly data collections.

    Args:
        data_collection: A HourlyContinuousCollection or HourlyDiscontinuousCollection
            which will be used to generate the hourly plot.
        legend_parameters: An optional LegendParameter object to change the display
            of the HourlyPlot (Default: None).
        base_point: An optional Point3D to be used as a starting point to generate
            the geometry of the plot (Default: (0, 0, 0)).
        x_dim: An optional number to set the X dimension of the mesh cells (Default: 1).
        y_dim: An optional number to set the Y dimension of the mesh cells (Default: 4).
        z_dim: An optional number to set the Z dimension of the entire chart. This
            will be used to make the colored_mesh3d of the chart vary in the Z
            dimension according to the data. The value input here should usually be
            several times larger than the x_dim or y_dim in order to be noticeable
            (e.g. 100). If 0, the colored_mesh3d will simply be flat. (Default: 0).
        reverse_y: Boolean to note whether the Y-axis of the chart is reversed
            such that time flows from the top of the chart to the bottom instead of
            the bottom to the top. (Default: False).

    Properties:
        * data_collection
        * legend_parameters
        * base_point
        * x_dim
        * y_dim
        * z_dim
        * reverse_y
        * colored_mesh2d
        * colored_mesh3d
        * legend
        * chart_border2d
        * chart_border3d
        * hour_lines2d
        * hour_lines3d
        * hour_label_points2d
        * hour_label_points3d
        * hour_labels
        * month_lines2d
        * month_lines3d
        * month_label_points2d
        * month_label_points3d
        * month_labels
        * title_text
        * lower_title_location
        * upper_title_location
        * analysis_period
        * values
        * colors
    """
    __slots__ = (
        '_data_collection', '_base_point', '_x_dim', '_y_dim', '_z_dim',
        '_num_y', '_num_x', '_container', '_hour_points', '_hour_text', '_hour_text_24',
        '_month_points', '_month_label_points', '_month_text', '_reverse_y')

    # editing HOUR_LABELS will change the labels produced for the entire chart
    HOUR_LABELS = (0, 6, 12, 18, 24)

    def __init__(self, data_collection, legend_parameters=None, base_point=Point3D(),
                 x_dim=1, y_dim=4, z_dim=0, reverse_y=False):
        """Initialize hourly plot."""
        # check the input objects
        acceptable_colls = (HourlyContinuousCollection, HourlyDiscontinuousCollection)
        assert isinstance(data_collection, acceptable_colls), 'HourlyPlot ' \
            'data_collection must be a HourlyContinuousCollection or ' \
            'HourlyDiscontinuousCollection. Got {}.'.format(type(data_collection))
        assert isinstance(base_point, Point3D), 'Expected Point3D for ' \
            'HourlyPlot base point. Got {}.'.format(type(base_point))

        # ensure the analysis period of the data collection has been validated
        if not data_collection.validated_a_period:
            data_collection = data_collection.validate_analysis_period()

        # assign the inputs as properties of this data collection
        self._data_collection = data_collection.to_immutable()
        self._base_point = base_point
        self._x_dim = self._check_dim(x_dim, 'x_dim')
        self._y_dim = self._check_dim(y_dim, 'y_dim')
        self._z_dim = 0
        if z_dim != 0:
            self._z_dim = self._check_dim(z_dim, 'z_dim')
        self._reverse_y = bool(reverse_y)

        # set properties to None that will be computed later
        self._hour_points = None
        self._hour_text = None
        self._month_points = None
        self._month_label_points = None
        self._month_text = None

        # create the graphic container from the inputs
        a_per = self.analysis_period
        if not a_per.is_reversed:
            self._num_x = a_per.end_time.doy - a_per.st_time.doy + 1
        else:
            num_doys = 365 if not a_per.is_leap_year else 366
            self._num_x = num_doys - a_per.st_time.doy + a_per.end_time.doy + 1
        if a_per.end_hour == 23:  # the last datetime is not included
            self._num_y = ((a_per.end_hour - a_per.st_hour + 1) * a_per.timestep)
        else:  # the last datetime is included
            if a_per.st_hour <= a_per.end_hour:
                self._num_y = ((a_per.end_hour - a_per.st_hour) * a_per.timestep) + 1
            else:
                self._num_y = 24 * a_per.timestep
        x_dist = x_dim * self._num_x
        y_dist = y_dim * self._num_y
        max_pt = Point3D(base_point.x + x_dist, base_point.y + y_dist, base_point.z)
        self._container = GraphicContainer(
            data_collection.values, base_point, max_pt, legend_parameters,
            data_collection.header.data_type, data_collection.header.unit)

    @classmethod
    def from_z_dim_per_unit(cls, data_collection, legend_parameters=None,
                            base_point=Point3D(), x_dim=1, y_dim=4, z_dim_per_unit=0):
        """Create HourlyPlot with an option to set the Z dimension on a per-unit basis.

        This is useful in cases of wanting to compare the Z dimensions of two plots
        to each other.

        Args:
            data_collection: A Hourly Collection, which will be used to generate
                the hourly plot.
            legend_parameters: An optional LegendParameter object to change the display
                of the HourlyPlot (Default: None).
            base_point: An optional Point3D to be used as a starting point to generate
                the geometry of the plot (Default: (0, 0, 0)).
            x_dim: Optional number to set the X dimension of the mesh cells (Default: 1).
            y_dim: Optional number to set the Y dimension of the mesh cells (Default: 4).
            z_dim: Optional number to set the Z dimension per unit of value in the input
                data_collection. If 0, the colored_mesh3d will be flat. (Default: 0).
        """
        z_dim = (data_collection.max - data_collection.min) * z_dim_per_unit
        return cls(data_collection, legend_parameters, base_point, x_dim, y_dim, z_dim)

    @property
    def data_collection(self):
        """The data collection assigned to this hourly plot."""
        return self._data_collection

    @property
    def legend_parameters(self):
        """The legend parameters customizing this hourly plot."""
        return self._container.legend_parameters

    @property
    def base_point(self):
        """Point3D for the base point of this hourly plot."""
        return self._base_point

    @property
    def x_dim(self):
        """A number for the X dimension of each cell of the hourly plot."""
        return self._x_dim

    @property
    def y_dim(self):
        """A number for the Y dimension of each cell of the hourly plot."""
        return self._y_dim

    @property
    def z_dim(self):
        """A number for the Z dimension of the entire hourly plot."""
        return self._z_dim

    @property
    def reverse_y(self):
        """Boolean to note whether the Y-axis of the chart is reversed.

        If True, time over the day flows from the top of the chart to the bottom
        instead of the bottom to the top.
        """
        return self._reverse_y

    @property
    def colored_mesh2d(self):
        """Get a colored Mesh2D for this graphic."""
        return self._compute_colored_mesh2d()

    @property
    def colored_mesh3d(self):
        """Get a colored Mesh3D for this graphic.

        Note that this will be the same as the colored_mesh2d if the z_dim
        value is 0.
        """
        return self._compute_colored_mesh3d()

    @property
    def legend(self):
        """The legend assigned to this graphic."""
        return self._container._legend

    @property
    def chart_border2d(self):
        """Get a Polyline2D for the border of the plot."""
        base_pt = Point2D(self._base_point.x, self._base_point.y)
        width = self._container.max_point.x - self._container.min_point.x
        height = self._container.max_point.y - self._container.min_point.y
        pgon = Polygon2D.from_rectangle(base_pt, Vector2D(0, 1), width, height)
        return Polyline2D.from_polygon(pgon)

    @property
    def chart_border3d(self):
        """Get a Polyline3D for the border of the plot."""
        plane = Plane(o=Point3D(0, 0, self._base_point.z))
        return Polyline3D.from_polyline2d(self.chart_border2d, plane)

    @property
    def hour_lines2d(self):
        """Get a list of LineSegment2D for the 6-hour intervals of the chart."""
        if not self._hour_points:
            self._compute_static_hour_line_pts()
        vec = Vector2D(self._num_x * self._x_dim)
        return [LineSegment2D(pt, vec) for pt in self._hour_points]

    @property
    def hour_lines3d(self):
        """Get a list of LineSegment3D for the 6-hour intervals of the chart."""
        if not self._hour_points:
            self._compute_static_hour_line_pts()
        vec = Vector3D(self._num_x * self._x_dim)
        return [LineSegment3D(Point3D(pt.x, pt.y, self._base_point.z), vec)
                for pt in self._hour_points]

    @property
    def hour_label_points2d(self):
        """Get a list of Point2Ds for the 6-hour text labels for the chart."""
        if not self._hour_points:
            self._compute_static_hour_line_pts()
        txt_hght = self.legend_parameters.text_height
        return [Point2D(pt.x - txt_hght * 2, pt.y) for pt in self._hour_points]

    @property
    def hour_label_points3d(self):
        """Get a list of Point3Ds for the 6-hour text labels for the chart."""
        if not self._hour_points:
            self._compute_static_hour_line_pts()
        txt_hght = self.legend_parameters.text_height
        return [Point3D(pt.x - txt_hght * 2, pt.y, self._base_point.z)
                for pt in self._hour_points]

    @property
    def hour_labels(self):
        """Get a list of text strings for the hour labels for the chart.

        These will be in 12-hour clock format.
        """
        if not self._hour_text:
            self._compute_static_hour_line_pts()
        return self._hour_text

    @property
    def hour_labels_24(self):
        """Get a list of text strings for the hour labels for the chart.

        These will be in 24-hour clock format.
        """
        if not self._hour_text_24:
            self._compute_static_hour_line_pts()
        return self._hour_text_24

    @property
    def month_lines2d(self):
        """Get a list of LineSegment2D for the month intervals of the chart."""
        if not self._month_points:
            self._compute_month_line_pts()
        vec = Vector2D(0, self._num_y * self._y_dim)
        return [LineSegment2D(pt, vec) for pt in self._month_points]

    @property
    def month_lines3d(self):
        """Get a list of LineSegment3D for the month intervals of the chart."""
        if not self._month_points:
            self._compute_month_line_pts()
        vec = Vector3D(0, self._num_y * self._y_dim, 0)
        return [LineSegment3D(Point3D(pt.x, pt.y, self._base_point.z), vec)
                for pt in self._month_points]

    @property
    def month_label_points2d(self):
        """Get a list of Point2Ds for the month text labels for the chart."""
        if not self._month_label_points:
            self._compute_month_line_pts()
        txt_hght = self.legend_parameters.text_height
        return [Point2D(pt.x, pt.y - txt_hght) for pt in self._month_label_points]

    @property
    def month_label_points3d(self):
        """Get a list of Point3Ds for the month text labels for the chart."""
        if not self._month_label_points:
            self._compute_month_line_pts()
        txt_hght = self.legend_parameters.text_height
        return [Point3D(pt.x, pt.y - txt_hght, self._base_point.z)
                for pt in self._month_label_points]

    @property
    def month_labels(self):
        """Get a list of text strings for the month labels for the chart."""
        if not self._month_text:
            self._compute_month_line_pts()
        return self._month_text

    @property
    def title_text(self):
        """Text string for the title of the hourly plot."""
        title_array = ['{} ({})'.format(self._data_collection.header.data_type,
                                        self._data_collection.header.unit)]
        title_array.append(str(self._data_collection.header.analysis_period))
        for key, val in self._data_collection.header.metadata.items():
            title_array.append('{}: {}'.format(key, val))
        return '\n'.join(title_array)

    @property
    def lower_title_location(self):
        """A Plane for the lower location of title text."""
        pln = self._container.lower_title_location
        txt_hght = self.legend_parameters.text_height
        return Plane(pln.n, Point3D(pln.o.x, pln.o.y - txt_hght * 2, pln.o.z))

    @property
    def upper_title_location(self):
        """A Plane for the upper location of title text."""
        num_lines = self.title_text.count('\n') + 4
        pln = self._container.upper_title_location
        txt_hght = self.legend_parameters.text_height
        return Plane(pln.n, Point3D(pln.o.x, pln.o.y + txt_hght * num_lines, pln.o.z))

    @property
    def analysis_period(self):
        """The AnalysisPeriod assigned to the hourly plot's data collection."""
        return self._data_collection.header.analysis_period

    @property
    def values(self):
        """A list of values assigned to this hourly plot.

        These will align correctly with the mesh faces, even when reverse_y is True.
        """
        if self._reverse_y:  # reverse each day of values before returning it
            rev_values = []
            dts = self._data_collection.datetimes
            day_values = []
            current_day = dts[0].day
            for dat_t, col in zip(dts, self._data_collection.values):
                if dat_t.day == current_day:
                    day_values.append(col)
                else:
                    day_values.reverse()
                    rev_values.extend(day_values)
                    current_day = dat_t.day
                    day_values = [col]
            day_values.reverse()
            rev_values.extend(day_values)
            return rev_values
        return self._data_collection.values

    @property
    def colors(self):
        """A list of colors assigned to the mesh faces of this hourly plot.

        These will align correctly with the mesh faces, even when reverse_y is True.
        """
        if self._reverse_y:  # reverse each day of colors before returning it
            rev_colors = []
            dts = self._data_collection.datetimes
            day_colors = []
            current_day = dts[0].day
            for dat_t, col in zip(dts, self._container.value_colors):
                if dat_t.day == current_day:
                    day_colors.append(col)
                else:
                    day_colors.reverse()
                    rev_colors.extend(day_colors)
                    current_day = dat_t.day
                    day_colors = [col]
            day_colors.reverse()
            rev_colors.extend(day_colors)
            return rev_colors
        return self._container.value_colors

    def custom_hour_lines2d(self, hour_labels):
        """Get a list of LineSegment2D for a list of numbers representing hour labels.

        Args:
            hour_labels: An array of numbers from 0 to 24 representing the hours
                to display. (eg. [0, 3, 6, 9, 12, 15, 18, 21, 24])
        """
        _hour_points, _hour_text, _hour_24 = self._compute_hour_line_pts(hour_labels)
        vec = Vector2D(self._num_x * self._x_dim)
        return [LineSegment2D(pt, vec) for pt in _hour_points]

    def custom_hour_lines3d(self, hour_labels):
        """Get a list of LineSegment3D for the 6-hour intervals of the chart.

        Args:
            hour_labels: An array of numbers from 0 to 24 representing the hours
                to display. (eg. [0, 3, 6, 9, 12, 15, 18, 21, 24])
        """
        _hour_points, _hour_text, _hour_24 = self._compute_hour_line_pts(hour_labels)
        vec = Vector3D(self._num_x * self._x_dim)
        return [LineSegment3D(Point3D(pt.x, pt.y, self._base_point.z), vec)
                for pt in _hour_points]

    def custom_hour_label_points2d(self, hour_labels):
        """Get a list of Point2Ds for the 6-hour text labels for the chart.

        Args:
            hour_labels: An array of numbers from 0 to 24 representing the hours
                to display. (eg. [0, 3, 6, 9, 12, 15, 18, 21, 24])
        """
        _hour_points, _hour_text, _hour_24 = self._compute_hour_line_pts(hour_labels)
        txt_hght = self.legend_parameters.text_height
        return [Point2D(pt.x - txt_hght * 2, pt.y) for pt in _hour_points]

    def custom_hour_label_points3d(self, hour_labels):
        """Get a list of Point3Ds for the 6-hour text labels for the chart.

        Args:
            hour_labels: An array of numbers from 0 to 24 representing the hours
                to display. (eg. [0, 3, 6, 9, 12, 15, 18, 21, 24])
        """
        _hour_points, _hour_text, _hour_24 = self._compute_hour_line_pts(hour_labels)
        txt_hght = self.legend_parameters.text_height
        return [Point3D(pt.x - txt_hght * 2, pt.y, self._base_point.z)
                for pt in _hour_points]

    def custom_hour_labels(self, hour_labels, clock_24=False):
        """Get a list of text strings for the 6-hour labels for the chart.

        Args:
            hour_labels: An array of numbers from 0 to 24 representing the hours
                to display. (eg. [0, 3, 6, 9, 12, 15, 18, 21, 24])
            clock_24: Boolean for whether a 24-hour clock should be used instead
                of the 12-hour clock.
        """
        _hour_points, _hour_text, _hour_24 = self._compute_hour_line_pts(hour_labels)
        return _hour_24 if clock_24 else _hour_text

    def _compute_colored_mesh2d(self):
        """Compute a colored mesh from this object's data collection."""
        # generate the base mesh as a standard grid
        _colored_mesh2d = Mesh2D.from_grid(
            self.base_point, self._num_x, self._num_y, self.x_dim, self.y_dim)

        # remove any faces in the base mesh that do not represent the data
        if not isinstance(self.data_collection, HourlyContinuousCollection):
            # get a pattern of booleans for whether mesh faces should be included
            data_coll_moys = [dt.moy for dt in self.data_collection.datetimes]
            data_coll_moys.append(527100)  # extra value for the end of the list
            found_i = 0
            mesh_pattern = []
            s_aper = self.analysis_period
            m_aper = s_aper if s_aper.st_hour <= s_aper.end_hour else \
                AnalysisPeriod(s_aper.st_month, s_aper.st_day, 0, s_aper.end_month,
                               s_aper.end_day, 23, s_aper.timestep, s_aper.is_leap_year)
            for moy in m_aper.moys:
                if moy == data_coll_moys[found_i]:
                    mesh_pattern.append(True)
                    found_i += 1
                else:
                    mesh_pattern.append(False)
            if self._reverse_y:
                hr_diff = abs(m_aper.end_hour - m_aper.st_hour)
                t_step = m_aper.timestep
                t_diff = t_step * hr_diff + 1 if t_step == 1 or hr_diff != 23 else \
                    t_step * (hr_diff + 1)
                mesh_pat_rev = []
                for i in range(0, len(mesh_pattern), t_diff):
                    mesh_pat_rev.extend(reversed(mesh_pattern[i:i + t_diff]))
                mesh_pattern = mesh_pat_rev
            _colored_mesh2d = _colored_mesh2d.remove_faces_only(mesh_pattern)

        # assign the colors to the mesh
        _colored_mesh2d.colors = self.colors
        return _colored_mesh2d

    def _compute_colored_mesh3d(self):
        """Compute a colored mesh from this object's data collection."""
        _colored_mesh3d = Mesh3D.from_mesh2d(
            self.colored_mesh2d, Plane(o=Point3D(0, 0, self._container.min_point.z)))
        if self.z_dim != 0:
            _colored_mesh3d = _colored_mesh3d.height_field_mesh(
                self.data_collection.values, (0, self.z_dim))
        return _colored_mesh3d

    def _compute_static_hour_line_pts(self):
        """Compute the points for the hour lines and labels."""
        self._hour_points, self._hour_text, self._hour_text_24 = \
            self._compute_hour_line_pts(self.HOUR_LABELS)

    def _compute_hour_line_pts(self, hour_labels):
        """Compute the points for the hour lines and labels."""
        s_aper = self.analysis_period
        m_aper = s_aper if s_aper.st_hour <= s_aper.end_hour else \
            AnalysisPeriod(s_aper.st_month, s_aper.st_day, 0, s_aper.end_month,
                           s_aper.end_day, 23, s_aper.timestep, s_aper.is_leap_year)
        st_hr = m_aper.st_hour
        end_hr = m_aper.end_hour + 1
        t_step = m_aper.timestep
        _hour_points = []
        _hour_text, _hour_text_24 = [], []
        last_hr = 0
        for hr in hour_labels:
            if st_hr <= hr <= end_hr:
                y_dist = (hr - st_hr) * self.y_dim * t_step
                pt_y = self.base_point.y + y_dist if not self._reverse_y else \
                    self._container.max_point.y - y_dist
                pt = Point2D(self._container.min_point.x, pt_y)
                _hour_points.append(pt)
                _hour_text_24.append('{}:00'.format(hr))
                hr_val = hr if hr <= 12 else hr - 12
                am_pm = 'PM' if 12 <= hr < 24 else 'AM'
                if hr_val == 0:
                    hr_val = 12
                _hour_text.append('{} {}'.format(hr_val, am_pm))
                last_hr = hr
        if self.analysis_period.timestep > 1 and last_hr == end_hr:
            _hour_points.pop(-1)
            _hour_text.pop(-1)
        return _hour_points, _hour_text, _hour_text_24

    def _compute_month_line_pts(self):
        """Compute the points for the month lines and labels."""
        # extract several reused properties from the analysis period
        st_mon = self.analysis_period.st_month
        end_mon = self.analysis_period.end_month
        st_day = self.analysis_period.st_day
        end_day = self.analysis_period.end_day
        dpm = AnalysisPeriod.NUMOFDAYSEACHMONTHLEAP \
            if self.analysis_period.is_leap_year \
            else AnalysisPeriod.NUMOFDAYSEACHMONTH

        # create a list of days in each of the months and collect month text
        self._month_text = [AnalysisPeriod.MONTHNAMES[st_mon]]
        days_list = []
        if st_mon == end_mon:
            days_list = [dpm[st_mon - 1] + 1 - st_day + (end_day - dpm[st_mon - 1])]
        else:
            days_list.append(dpm[st_mon - 1] + 1 - st_day)
        month_count = len(self.analysis_period.months_int) - 2
        for count, month in enumerate(self.analysis_period.months_int[1:]):
            self._month_text.append(AnalysisPeriod.MONTHNAMES[month])
            if count == month_count:
                days_list.append(end_day)
            else:
                days_list.append(dpm[month - 1])

        # create the points for each of the month labels
        b_pt = self._base_point
        self._month_label_points = \
            [Point2D(b_pt.x + (days_list[0] / 2) * self.x_dim, b_pt.y)]
        for mon_days in days_list[1:]:
            prev_pt = self._month_label_points[-1]
            new_pt = Point2D(prev_pt.x + mon_days * self.x_dim, prev_pt.y)
            self._month_label_points.append(new_pt)

        # create the points for each of the month lines
        self._month_points = []
        if len(days_list) > 1:
            prev_x = b_pt.x + days_list[0] * self.x_dim
            for mon_days in days_list[1:]:
                self._month_points.append(Point2D(prev_x, b_pt.y))
                prev_x += mon_days * self.x_dim

    @staticmethod
    def _check_dim(dim_value, dim_name):
        """Check a given value for a dimension input."""
        assert isinstance(dim_value, (float, int)), 'Expected number for ' \
            'HourlyPlot {}. Got {}.'.format(dim_name, type(dim_value))
        assert dim_value > 0, 'HourlyPlot {} must be greater than 0. ' \
            'Got {}.'.format(dim_name, dim_value)
        return dim_value

    def __repr__(self):
        """Hourly Plot representation."""
        return 'Hourly Plot:\n{}'.format(self.data_collection.header)



# coding=utf-8
"""Resistivity data type."""
from __future__ import division

from .base import DataTypeBase


class Resistivity(DataTypeBase):
    """Resistivity
    """
    _units = ('K-m/W', 'F-ft-h/Btu')
    _si_units = ('K-m/W',)
    _ip_units = ('F-ft-h/Btu',)
    _min = 0
    _abbreviation = 'rho'

    def _Km_W_to_Ffth_Btu(self, value):
        return value * 1.7295772056

    def _Ffth_Btu_to_Km_W(self, value):
        return value / 1.7295772056

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('K-m/W', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        else:
            return self.to_unit(values, 'F-ft-h/Btu', from_unit), 'F-ft-h/Btu'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        else:
            return self.to_unit(values, 'K-m/W', from_unit), 'K-m/W'



# coding=utf-8
"""Distance data type."""
from __future__ import division

from .base import DataTypeBase


class Distance(DataTypeBase):
    """Distance
    """
    _units = ('m', 'ft', 'mm', 'in', 'km', 'mi', 'cm')
    _si_units = ('m', 'mm', 'km', 'cm')
    _ip_units = ('ft', 'in', 'mi')
    _min = 0
    _cumulative = True
    _abbreviation = 'D'

    def _m_to_ft(self, value):
        return value * 3.28084

    def _m_to_mm(self, value):
        return value * 1000.

    def _m_to_in(self, value):
        return value * 39.3701

    def _m_to_km(self, value):
        return value / 1000.

    def _m_to_mi(self, value):
        return value / 1609.344

    def _m_to_cm(self, value):
        return value * 100.

    def _ft_to_m(self, value):
        return value / 3.28084

    def _mm_to_m(self, value):
        return value / 1000.

    def _in_to_m(self, value):
        return value / 39.3701

    def _km_to_m(self, value):
        return value * 1000.

    def _mi_to_m(self, value):
        return value * 1609.344

    def _cm_to_m(self, value):
        return value / 100.

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('m', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        elif from_unit == 'mm':
            return self.to_unit(values, 'in', from_unit), 'in'
        elif from_unit == 'km':
            return self.to_unit(values, 'mi', from_unit), 'mi'
        else:
            return self.to_unit(values, 'ft', from_unit), 'ft'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        elif from_unit == 'in':
            return self.to_unit(values, 'mm', from_unit), 'mm'
        elif from_unit == 'mi':
            return self.to_unit(values, 'km', from_unit), 'km'
        else:
            return self.to_unit(values, 'm', from_unit), 'm'


class Visibility(Distance):
    """Visibility
    """
    _abbreviation = 'Vis'
    _missing_epw = 9999


class CeilingHeight(Distance):
    """CeilingHeight
    """
    _abbreviation = 'Hciel'
    _missing_epw = 99999


class PrecipitableWater(Distance):
    """PrecipitableWater
    """
    _abbreviation = 'PW'
    _missing_epw = 999


class SnowDepth(Distance):
    """SnowDepth
    """
    _abbreviation = 'Dsnow'
    _missing_epw = 999


class LiquidPrecipitationDepth(Distance):
    """LiquidPrecipitationDepth
    """
    _abbreviation = 'LPD'
    _missing_epw = 999



# coding=utf-8
"""Module for visualizing and converting between meteorological and other wind speeds."""
from __future__ import division
import math

from ladybug_geometry.geometry3d import Vector3D, Point3D, Plane, LineSegment3D, \
    Polyline3D, Mesh3D
from .datacollection import HourlyContinuousCollection
from .color import Color


class WindProfile(object):
    """Object for visualizing and converting from meteorological wind speeds.

    Args:
        terrain: A text string that sets the terrain class associated with the
            location that the wind profile represents. (Default: city). Values
            must be one the following:

            * city - 50% of buildings above 21m over a distance of at least 2000m upwind.
            * suburban - suburbs, wooded areas.
            * country - open, with scattered objects generally less than 10m high.
            * water - flat areas downwind of a large water body (max 500m inland).

        meteorological_terrain: A text string that sets the terrain class associated
            with the meteorological wind speed. (Default: country, which is typical
            of most airports where wind measurements are taken).
        meteorological_height: A number for the height above the ground at which
            the meteorological wind speed is measured in meters. (Default: 10 meters).
        log_law: A boolean to note whether the wind profile should use a logarithmic
            law to determine wind speeds instead of the default power law, which
            is used by EnergyPlus. (Default: False).

    Properties:
        * terrain
        * meteorological_terrain
        * meteorological_height
        * log_law
        * boundary_layer_height
        * power_law_exponent
        * roughness_length
        * met_boundary_layer_height
        * met_power_law_exponent
        * met_roughness_length
    """
    __slots__ = (
        '_terrain', '_meteorological_terrain', '_meteorological_height', '_log_law',
        '_boundary_layer_height', '_power_law_exponent', '_roughness_length',
        '_met_boundary_layer_height', '_met_power_law_exponent', '_met_roughness_length',
        '_met_power_denom', '_met_log_denom')

    TERRAINS = ('city', 'suburban', 'country', 'water')
    TERRAIN_PARAMETERS = {
        'city': (460, 0.33, 1.0),
        'suburban': (370, 0.22, 0.5),
        'country': (270, 0.14, 0.1),
        'water': (210, 0.10, 0.03)
    }
    BLACK = Color(0, 0, 0)

    def __init__(self, terrain='city', meteorological_terrain='country',
                 meteorological_height=10, log_law=False):
        """Initialize wind profile."""
        self._meteorological_height = 10  # set a default to ensure checks pass
        self.terrain = terrain
        self.meteorological_terrain = meteorological_terrain
        self.meteorological_height = meteorological_height
        self.log_law = log_law

    @property
    def terrain(self):
        """Get or set text for the terrain class for the wind profile location.

        Setting this will set all of the properties of the boundary layer,
        roughness length, etc. Choose from the following options.

        * city
        * suburban
        * country
        * water
        """
        return self._terrain

    @terrain.setter
    def terrain(self, value):
        value = self._check_terrain(value)
        self._terrain = value
        self._boundary_layer_height, self._power_law_exponent, self._roughness_length = \
            self.TERRAIN_PARAMETERS[value]

    @property
    def meteorological_terrain(self):
        """Get or set text for the terrain class for the meteorological location.

        Setting this will set all of the properties of the boundary layer,
        roughness length, etc. Choose from the following options.

        * city
        * suburban
        * country
        * water
        """
        return self._meteorological_terrain

    @meteorological_terrain.setter
    def meteorological_terrain(self, value):
        value = self._check_terrain(value)
        self._meteorological_terrain = value
        self._met_boundary_layer_height, self._met_power_law_exponent, \
            self._met_roughness_length = self.TERRAIN_PARAMETERS[value]
        self._compute_met_power_denom()
        self._compute_met_log_denom()

    @property
    def meteorological_height(self):
        """Get or set the measurement height of the meteorological wind speed [m]."""
        return self._meteorological_height

    @meteorological_height.setter
    def meteorological_height(self, value):
        assert isinstance(value, (float, int)), 'Expected number for ' \
            'WindProfile meteorological_height. Got {}.'.format(type(value))
        assert value > 0, 'WindProfile meteorological_height must be ' \
            'greater than 0. Got {}.'.format(value)
        self._meteorological_height = value
        self._compute_met_power_denom()
        self._compute_met_log_denom()

    @property
    def log_law(self):
        """A boolean to note whether the wind profile should be using a logarithmic law.
        """
        return self._log_law

    @log_law.setter
    def log_law(self, value):
        self._log_law = bool(value)

    @property
    def boundary_layer_height(self):
        """Get or set the boundary layer height of the wind profile location [m]."""
        return self._boundary_layer_height

    @boundary_layer_height.setter
    def boundary_layer_height(self, value):
        assert isinstance(value, (float, int)), 'Expected number for ' \
            'WindProfile boundary_layer_height. Got {}.'.format(type(value))
        assert value > 0, 'WindProfile boundary_layer_height must be ' \
            'greater than 0. Got {}.'.format(value)
        self._boundary_layer_height = value

    @property
    def power_law_exponent(self):
        """Get or set the power law exponent of the wind profile location."""
        return self._power_law_exponent

    @power_law_exponent.setter
    def power_law_exponent(self, value):
        assert isinstance(value, (float, int)), 'Expected number for ' \
            'WindProfile power_law_exponent. Got {}.'.format(type(value))
        assert 1 > value > 0, 'WindProfile power_law_exponent must be ' \
            'between 0 and 1. Got {}.'.format(value)
        self._power_law_exponent = value

    @property
    def roughness_length(self):
        """Get or set the roughness length of the wind profile location [m]."""
        return self._roughness_length

    @roughness_length.setter
    def roughness_length(self, value):
        assert isinstance(value, (float, int)), 'Expected number for ' \
            'WindProfile roughness_length. Got {}.'.format(type(value))
        assert value > 0, 'WindProfile roughness_length must be ' \
            'greater than 0. Got {}.'.format(value)
        self._roughness_length = value

    @property
    def met_boundary_layer_height(self):
        """Get or set the boundary layer height of the meteorological location [m]."""
        return self._met_boundary_layer_height

    @met_boundary_layer_height.setter
    def met_boundary_layer_height(self, value):
        assert isinstance(value, (float, int)), 'Expected number for ' \
            'WindProfile met_boundary_layer_height. Got {}.'.format(type(value))
        assert value > 0, 'WindProfile met_boundary_layer_height must be ' \
            'greater than 0. Got {}.'.format(value)
        self._met_boundary_layer_height = value
        self._compute_met_power_denom()

    @property
    def met_power_law_exponent(self):
        """Get or set the power law exponent of the meteorological location."""
        return self._met_power_law_exponent

    @met_power_law_exponent.setter
    def met_power_law_exponent(self, value):
        assert isinstance(value, (float, int)), 'Expected number for ' \
            'WindProfile met_power_law_exponent. Got {}.'.format(type(value))
        assert 1 > value > 0, 'WindProfile met_power_law_exponent must be ' \
            'between 0 and 1. Got {}.'.format(value)
        self._met_power_law_exponent = value
        self._compute_met_power_denom()

    @property
    def met_roughness_length(self):
        """Get or set the roughness length of the meteorological location [m]."""
        return self._met_roughness_length

    @met_roughness_length.setter
    def met_roughness_length(self, value):
        assert isinstance(value, (float, int)), 'Expected number for ' \
            'WindProfile met_roughness_length. Got {}.'.format(type(value))
        assert value > 0, 'WindProfile met_roughness_length must be ' \
            'greater than 0. Got {}.'.format(value)
        self._met_roughness_length = value
        self._compute_met_log_denom()

    def calculate_wind(self, meteorological_wind_speed, height=1):
        """Calculate the wind speed at a given height above the ground.

        Args:
            meteorological_wind_speed: A number for the meteorological
                wind speed [m/s].
            height: The height above the ground to be evaluated in
                meters. (Default: 1).
        """
        if self._log_law:
            if height > self._roughness_length:
                met_log_num = math.log(height / self._roughness_length)
                return meteorological_wind_speed * (met_log_num / self._met_log_denom)
            return 0
        else:
            h_ratio = (height / self._boundary_layer_height) ** self._power_law_exponent
            return h_ratio * (meteorological_wind_speed * self._met_power_denom)

    def calculate_wind_data(self, meteorological_wind_data, height=1):
        """Get a data collection of wind speed at a given height above the ground.

        Args:
            meteorological_wind_data: A data collection of meteorological
                wind speed [m/s].
            height: The height above the ground to be evaluated in
                meters. (Default: 1).
        """
        vals = tuple(self.calculate_wind(v, height) for v in meteorological_wind_data)
        new_header = meteorological_wind_data.header.duplicate()
        new_header.metadata['height'] = '{}m'.format(round(height, 1))
        if isinstance(meteorological_wind_data, HourlyContinuousCollection):
            return HourlyContinuousCollection(new_header, vals)
        else:
            dts = meteorological_wind_data.datetimes
            return meteorological_wind_data.__class__(new_header, vals, dts)

    def wind_vector(
            self, meteorological_wind_speed, height, direction=None,
            length_dimension=1, scale_factor=1):
        """Get a Vector3D for a wind profile arrow at a given height above the ground.

        Args:
            meteorological_wind_speed: A number for the meteorological wind speed [m/s].
            height: The height above the ground to be evaluated in meters.
            direction: An optional number between 0 and 360 that represents the
                cardinal direction that the wind vector is facing in the XY
                plane. 0 = North, 90 = East, 180 = South, 270 = West. If None,
                the wind vector will simply be placed in the XY plane. (Default: None).
            length_dimension: A number to denote the length dimension of a 1 m/s
                wind vector in meters. This will be used to set the length of the
                wind vector. (Default: 1).
            scale_factor: An optional number that will be multiplied by all dimensions
                to account for the fact that the wind profile may be displaying in
                a units system other than meters. (Default: 1).

        Returns:
            A ladybug-geometry Vector3D representing the wind vector.
        """
        direction = self._flip_direction(direction)
        wind_speed = self.calculate_wind(meteorological_wind_speed, height)
        vec_mag = wind_speed * length_dimension * scale_factor
        if direction is None:
            return Vector3D(vec_mag, 0, 0)
        else:
            base_vec = Vector3D(0, vec_mag, 0)
            return base_vec.rotate_xy(-math.radians(direction))

    def profile_polyline3d(
            self, meteorological_wind_speed, max_height=30, vector_spacing=2,
            direction=None, base_point=Point3D(0, 0, 0), length_dimension=5,
            scale_factor=1):
        """Get a Polyline3D for a wind profile curve at a meteorological wind speed.

        Args:
            meteorological_wind_speed: A number for the meteorological wind speed [m/s].
            max_height: A number in meters to specify the maximum height of the
                wind profile curve. (Default: 30 meters).
            vector_spacing: A number in meters to specify the difference in height
                between each of the wind vectors that is used to build the
                profile curve. Lower numbers will result in smoother looking
                curves. (Default 2 meters).
            direction: An optional number between 0 and 360 that represents the
                cardinal direction that the wind profile is facing in the XY
                plane. 0 = North, 90 = East, 180 = South, 270 = West. If None,
                the wind profile will simply be placed in the XY plane. (Default: None).
            base_point: A ladybug-geometry Point3D that represents the ground
                location of the wind profile. (Default, (0, 0, 0)).
            length_dimension: A number to denote the length dimension of a 1 m/s
                wind vector in meters. (Default: 5).
            scale_factor: An optional number that will be multiplied by all dimensions
                to account for the fact that the wind profile may be displaying in
                a units system other than meters. (Default: 1).

        Returns:
            A tuple with three values.

            - profile_polyline: A ladybug-geometry Polyline3D representing the
                wind profile.

            - wind_vectors: A list of ladybug-geometry Vector3D representing the
                wind vectors that built the profile.

            - anchor_pts: A list of ladybug-geometry Point3D representing the
                anchor points for the wind vectors.
        """
        self._check_profile_inputs(max_height, vector_spacing)
        bp = base_point
        profile_pts, wind_vectors, anchor_pts = [bp], [Vector3D(0, 0, 0)], [bp]
        m_val = max_height + vector_spacing
        for h in self._frange(vector_spacing, m_val, vector_spacing):
            a_pt = Point3D(bp.x, bp.y + (h * scale_factor), bp.z) \
                if direction is None else Point3D(bp.x, bp.y, bp.z + (h * scale_factor))
            w_vec = self.wind_vector(
                meteorological_wind_speed, h, direction,
                length_dimension, scale_factor)
            profile_pts.append(a_pt.move(w_vec))
            wind_vectors.append(w_vec)
            anchor_pts.append(a_pt)
        profile_polyline = Polyline3D(profile_pts)
        return profile_polyline, wind_vectors, anchor_pts

    def mesh_arrow(
            self, meteorological_wind_speed, height, direction=None,
            base_point=Point3D(0, 0, 0), length_dimension=5, height_dimension=1,
            scale_factor=1):
        """Get a Mesh3D for an arrow at a given height above the ground.

        Args:
            meteorological_wind_speed: A number for the meteorological wind speed [m/s].
            height: The height above the ground to be evaluated in meters.
            direction: An optional number between 0 and 360 that represents the
                cardinal direction that the mesh arrow is facing in the XY
                plane. 0 = North, 90 = East, 180 = South, 270 = West. If None,
                the wind vector and mesh arrow will simply be placed in the XY
                plane. (Default: None).
            base_point: A ladybug-geometry Point3D that represents the ground
                location of the wind profile. (Default, (0, 0, 0)).
            length_dimension: A number to denote the length dimension of a 1 m/s
                wind vector in meters. This will be used to set the length of the
                wind vector arrow. (Default: 5).
            height_dimension: A number to denote the height dimension of the
                wind vector in meters. (Default: 1).
            scale_factor: An optional number that will be multiplied by all dimensions
                to account for the fact that the wind profile may be displaying in
                a units system other than meters. (Default: 1).

        Returns:
            A tuple with four values.

            - mesh_arrow: A Mesh3D object that represents the wind speed at the
                height above the ground.

            - wind_vector: A ladybug-geometry Vector3D representing the wind vector.

            - anchor_pt: A ladybug-geometry Point3D representing the anchor point
                for the wind vector.

            - wind_speed: A number for the wind speed associated with the mesh arrow.
        """
        direction = self._flip_direction(direction)
        # get an anchor point that represents the wind arrow
        bp = base_point
        if direction is None:
            y_val = bp.y + height * scale_factor
            anchor_pt = Point3D(bp.x, y_val, bp.z)
        else:
            z_val = bp.z + height * scale_factor
            anchor_pt = Point3D(bp.x, bp.y, z_val)
        # calculate the wind speed and vector at the height
        wind_speed = self.calculate_wind(meteorological_wind_speed, height)
        vec_mag = wind_speed * length_dimension
        vm_scaled = vec_mag * scale_factor
        # build the mesh vertices and with arrow facing the X-axis from the anchor point
        x_vec = Vector3D(vm_scaled, 0, 0)
        max_arrow = length_dimension / 2
        if vec_mag >= length_dimension:
            ar_vec = Vector3D(scale_factor * max_arrow, 0, 0)
            ar_x = (vec_mag - max_arrow) * scale_factor
        else:
            mid_val = ar_x = vm_scaled * 0.5
            ar_vec = Vector3D(mid_val, 0, 0)
        if direction is None:
            ar_pt = Point3D(bp.x + ar_x, bp.y + (height * scale_factor), bp.z)
        else:
            ar_pt = Point3D(bp.x + ar_x, bp.y, bp.z + (height * scale_factor))
        ad = (height_dimension * scale_factor) / 2
        hd = (height_dimension * scale_factor) / 4
        stem_mvs = (Vector3D(0, hd, hd), Vector3D(0, -hd, hd),
                    Vector3D(0, -hd, -hd), Vector3D(0, hd, -hd))
        ar_mvs = (Vector3D(0, ad, ad), Vector3D(0, -ad, ad),
                  Vector3D(0, -ad, -ad), Vector3D(0, ad, -ad))
        verts = []
        for mv in stem_mvs + (x_vec,):
            verts.append(anchor_pt.move(mv))
        for mv in ar_mvs + (ar_vec,):
            verts.append(ar_pt.move(mv))
        # rotate the vertices if the direction is specified
        if direction is not None:
            v_ang, rad90 = -math.radians(direction), math.radians(90)
            verts = [v.rotate_xy(v_ang + rad90, base_point) for v in verts]
            base_vec = Vector3D(0, vec_mag, 0)
            wind_vector = base_vec.rotate_xy(v_ang)
        else:
            wind_vector = Vector3D(vec_mag, 0, 0)
        # specify the faces and return the final Mesh3D
        faces = [
            (0, 1, 2, 3), (0, 1, 4), (1, 2, 4), (2, 3, 4), (3, 0, 4),
            (5, 6, 7, 8), (5, 6, 9), (6, 7, 9), (7, 8, 9), (8, 5, 9)
        ]
        mesh_arrow = Mesh3D(verts, faces)
        return mesh_arrow, wind_vector, anchor_pt, wind_speed

    def mesh_arrow_profile(
            self, meteorological_wind_speed, max_height=30, vector_spacing=2,
            direction=None, base_point=Point3D(0, 0, 0), length_dimension=5,
            height_dimension=1, scale_factor=1):
        """Get a Polyline3D for a wind profile curve at a meteorological wind speed.

        Args:
            meteorological_wind_speed: A number for the meteorological wind speed [m/s].
            max_height: A number in meters to specify the maximum height of the
                wind profile curve. (Default: 30 meters).
            vector_spacing: A number in meters to specify the difference in height
                between each of the mesh arrows. (Default 2 meters).
            direction: An optional number between 0 and 360 that represents the
                cardinal direction that the wind profile is facing in the XY
                plane. 0 = North, 90 = East, 180 = South, 270 = West. If None,
                the wind profile will simply be placed in the XY plane. (Default: None).
            base_point: A ladybug-geometry Point3D that represents the ground
                location of the wind profile. (Default, (0, 0, 0)).
            length_dimension: A number to denote the length dimension of a 1 m/s
                wind vector in meters. (Default: 5).
            height_dimension: A number to denote the height dimension of the
                wind vector in meters. (Default: 1).
            scale_factor: An optional number that will be multiplied by all dimensions
                to account for the fact that the wind profile may be displaying in
                a units system other than meters. (Default: 1).

        Returns:
            A tuple with five values

            - profile_polyline: A ladybug-geometry Polyline3D representing the
                wind profile.

            - mesh_arrows: A list of ladybug-geometry Mesh3D objects that
                represent the wind speed along with wind profile.

            - wind_speeds: A list of numbers for the wind speed associated with
                the mesh arrows.

            - wind_vectors: A list of ladybug-geometry Vector3D representing the
                wind vectors that built the profile.

            - anchor_pts: A list of ladybug-geometry Point3D representing the
                anchor points for the wind vectors.
        """
        self._check_profile_inputs(max_height, vector_spacing)
        bp = base_point
        mesh_arrows, wind_speeds = [], []
        profile_pts, wind_vectors, anchor_pts = [bp], [Vector3D(0, 0, 0)], [bp]
        m_val = max_height + vector_spacing
        for h in self._frange(vector_spacing, m_val, vector_spacing):
            m_ar, w_vec, a_pt, ws = self.mesh_arrow(
                meteorological_wind_speed, h, direction,
                base_point, length_dimension, height_dimension, scale_factor)
            mesh_arrows.append(m_ar)
            wind_speeds.append(ws)
            profile_pts.append(a_pt.move(w_vec))
            wind_vectors.append(w_vec)
            anchor_pts.append(a_pt)
        profile_polyline = Polyline3D(profile_pts)
        return profile_polyline, mesh_arrows, wind_speeds, wind_vectors, anchor_pts

    def speed_axis(
            self, max_speed=5, direction=None, base_point=Point3D(0, 0, 0),
            length_dimension=5, scale_factor=1, text_height=None):
        """Get a several objects for representing the X axis for wind speed.

        Args:
            max_speed: A number for the maximum wind speed along the axis
                in [m/s]. (Default: 5)
            direction: An optional number between 0 and 360 that represents the
                cardinal direction that the axis is facing in the XY
                plane. 0 = North, 90 = East, 180 = South, 270 = West. If None,
                the axis will simply be placed in the XY plane. (Default: None).
            base_point: A ladybug-geometry Point3D that represents the origin
                of the axis. (Default, (0, 0, 0)).
            length_dimension: A number to denote the length dimension of a 1 m/s
                wind vector in meters. (Default: 5).
            scale_factor: An optional number that will be multiplied by all dimensions
                to account for the fact that the wind profile may be displaying in
                a units system other than meters. (Default: 1).
            text_height: An optional number for the height of the text in the axis
                label. If None, the text height will be inferred based on the
                length_dimension. (Default: None).

        Returns:
            A tuple with three values.

            - axis_line: A ladybug-geometry Linesegment3D representing the axis.

            - axis_arrow: A ladybug-geometry Mesh3D representing the arrow head.

            - axis_ticks: A list of ladybug-geometry LineSegment3D representing the
                marks of speeds along the axis.

            - text_planes: A list of ladybug-geometry Planes for the axis text labels.

            - text: A list of text strings tha align with the text_planes for the
                text to display in the 3D scene.
        """
        # construct the axis line along the world X axis
        direction = self._flip_direction(direction)
        bp, max_speed = base_point, int(max_speed)
        end_x = bp.x + ((max_speed + 1) * length_dimension * scale_factor)
        ax_end_pt = Point3D(end_x, bp.y, bp.z)
        axis_line = LineSegment3D.from_end_points(bp, ax_end_pt)
        step_d = length_dimension * scale_factor
        # create the arrow at the end of the axis as a mesh
        a_d = step_d / 8
        ea_pt = Point3D(end_x + a_d * 3, bp.y, bp.z)
        if direction is None:
            m_pts = (Point3D(end_x, bp.y + a_d, bp.z),
                     Point3D(end_x, bp.y - a_d, bp.z), ea_pt)
        else:
            m_pts = (Point3D(end_x, bp.y, bp.z + a_d),
                     Point3D(end_x, bp.y, bp.z - a_d), ea_pt)
        axis_arrow = Mesh3D(m_pts, ((0, 1, 2),), colors=(self.BLACK,))
        # create the axis ticks and text planes
        tick_d = step_d / 6
        axis_ticks, text_planes, text = [], [], []
        for i in range(1, max_speed + 1):
            pt_x = bp.x + (step_d * i)
            pt_1 = Point3D(pt_x, bp.y, bp.z)
            if direction is None:
                pt_2 = Point3D(pt_x, bp.y - tick_d, bp.z)
                pt_3 = Point3D(pt_x, bp.y - (tick_d * 2), bp.z)
                txt_p = Plane(n=Vector3D(0, 0, 1), o=pt_3, x=Vector3D(1, 0, 0))
            else:
                pt_2 = Point3D(pt_x, bp.y, bp.z - tick_d)
                pt_3 = Point3D(pt_x, bp.y, bp.z - (tick_d * 2))
                txt_p = Plane(n=Vector3D(0, -1, 0), o=pt_3, x=Vector3D(1, 0, 0))
            axis_ticks.append(LineSegment3D.from_end_points(pt_1, pt_2))
            text_planes.append(txt_p)
            text.append(str(i))
        # add a text plane and string for the axis title
        txt_height = length_dimension / 2 if text_height is None else text_height
        sub_d = (tick_d * 2) + (txt_height * 2)
        ti_x = axis_line.midpoint.x
        if direction is None:
            pt_4 = Point3D(ti_x, bp.y - sub_d, bp.z)
            text_planes.append(Plane(n=Vector3D(0, 0, 1), o=pt_4, x=Vector3D(1, 0, 0)))
        else:
            pt_4 = Point3D(ti_x, bp.y, bp.z - sub_d)
            text_planes.append(Plane(n=Vector3D(0, -1, 0), o=pt_4, x=Vector3D(1, 0, 0)))
        text.append('Wind Speed (m/s)')
        # rotate the whole axis if a direction is specified
        if direction is not None:
            dir_radians = math.radians(90) - math.radians(direction)
            axis_line = axis_line.rotate_xy(dir_radians, bp)
            axis_arrow = axis_arrow.rotate_xy(dir_radians, bp)
            axis_ticks = [t.rotate_xy(dir_radians, bp) for t in axis_ticks]
            text_planes = [p.rotate_xy(dir_radians, bp) for p in text_planes]
        return axis_line, axis_arrow, axis_ticks, text_planes, text

    def height_axis(
            self, max_height=30, tick_spacing=4, direction=None,
            base_point=Point3D(0, 0, 0), scale_factor=1, text_height=None,
            feet_labels=False):
        """Get a several objects for representing the Y axis for height above the ground.

        Args:
            max_height: A number in meters to specify the maximum height of the
                wind profile curve. (Default: 30 meters).
            tick_spacing: A number in meters to specify the difference between
                each of axis ticks. (Default 4 meters).
            direction: An optional number between 0 and 360 that represents the
                cardinal direction that the wind profile is facing in the XY
                plane. 0 = North, 90 = East, 180 = South, 270 = West. If None,
                the axis will simply be placed in the XY plane. (Default: None).
            base_point: A ladybug-geometry Point3D that represents the ground
                location of the wind profile. (Default, (0, 0, 0)).
            scale_factor: An optional number that will be multiplied by all dimensions
                to account for the fact that the wind profile may be displaying in
                a units system other than meters. (Default: 1).
            text_height: An optional number for the height of the text in the axis
                label. If None, the text height will be inferred based on the
                tick_spacing. (Default: None).
            feet_labels: A boolean to note whether the text labels should be in
                feet (True) or meters (False). (Default: False).

        Returns:
            A tuple with three values.

            - axis_line: A ladybug-geometry Linesegment3D representing the axis.

            - axis_arrow: A ladybug-geometry Mesh3D representing the arrow head.

            - axis_ticks: A list of ladybug-geometry LineSegment3D representing the
                marks of heights along the axis.

            - text: A list of text strings tha align with the text_planes for the
                text to display in the 3D scene.
        """
        # construct the axis line
        direction = self._flip_direction(direction)
        bp = base_point
        end_v = (max_height + (tick_spacing / 2)) * scale_factor
        ax_end_pt = Point3D(bp.x, bp.y + end_v, bp.z) if direction is None \
            else Point3D(bp.x, bp.y, bp.z + end_v)
        axis_line = LineSegment3D.from_end_points(bp, ax_end_pt)
        step_d = tick_spacing * scale_factor
        # create the arrow at the end of the axis as a mesh
        a_d = step_d / 10
        if direction is None:
            ea_pt = Point3D(bp.x, ax_end_pt.y + (a_d * 3), bp.z)
            m_pts = (Point3D(bp.x + a_d, ax_end_pt.y, bp.z),
                     Point3D(bp.x - a_d, ax_end_pt.y, bp.z), ea_pt)
        else:
            ea_pt = Point3D(bp.x, bp.y, ax_end_pt.z + (a_d * 3))
            m_pts = (Point3D(bp.x + a_d, bp.y, ax_end_pt.z),
                     Point3D(bp.x - a_d, bp.y, ax_end_pt.z), ea_pt)
        axis_arrow = Mesh3D(m_pts, ((0, 1, 2),), colors=(self.BLACK,))
        # create the axis ticks and text planes
        tick_d = step_d / 6
        axis_ticks, text_planes, text = [], [], []
        mh = max_height + tick_spacing if max_height % tick_spacing == 0 else max_height
        for i in self._frange(0, mh, tick_spacing):
            pt_v = i * scale_factor
            if direction is None:
                pt_1 = Point3D(bp.x, bp.y + pt_v, bp.z)
                pt_2 = Point3D(bp.x - tick_d, bp.y + pt_v, bp.z)
                pt_3 = Point3D(bp.x - (tick_d * 2), bp.y + pt_v, bp.z)
                txt_p = Plane(n=Vector3D(0, 0, 1), o=pt_3, x=Vector3D(1, 0, 0))
            else:
                pt_1 = Point3D(bp.x, bp.y, bp.z + pt_v)
                pt_2 = Point3D(bp.x - tick_d, bp.y, bp.z + pt_v)
                pt_3 = Point3D(bp.x - (tick_d * 2), bp.y, bp.z + pt_v)
                txt_p = Plane(n=Vector3D(0, -1, 0), o=pt_3, x=Vector3D(1, 0, 0))
            axis_ticks.append(LineSegment3D.from_end_points(pt_1, pt_2))
            text_planes.append(txt_p)
            txt_str = str(int(i)) if not feet_labels else str(int(i * 3.28084))
            text.append(txt_str)
        # add a text plane and string for the axis title
        txt_height = tick_spacing / 2 if text_height is None else text_height
        sub_d = (tick_d * 2) + (txt_height * 4)
        if direction is None:
            ti_v = axis_line.midpoint.y
            pt_4 = Point3D(bp.x - sub_d, ti_v, bp.z)
            text_planes.append(Plane(n=Vector3D(0, 0, 1), o=pt_4, x=Vector3D(0, 1, 0)))
        else:
            ti_v = axis_line.midpoint.z
            pt_4 = Point3D(bp.x - sub_d, bp.y, ti_v)
            text_planes.append(Plane(n=Vector3D(0, -1, 0), o=pt_4, x=Vector3D(0, 0, 1)))
        units = 'ft' if feet_labels else 'm'
        text.append('Height ({})'.format(units))
        # rotate the whole axis if a direction is specified
        if direction is not None:
            dir_radians = math.radians(90) - math.radians(direction)
            axis_line = axis_line.rotate_xy(dir_radians, bp)
            axis_arrow = axis_arrow.rotate_xy(dir_radians, bp)
            axis_ticks = [t.rotate_xy(dir_radians, bp) for t in axis_ticks]
            text_planes = [p.rotate_xy(dir_radians, bp) for p in text_planes]
        return axis_line, axis_arrow, axis_ticks, text_planes, text

    def legend_plane(
            self, max_speed=5, direction=None, base_point=Point3D(0, 0, 0),
            length_dimension=5, scale_factor=1):
        """Get a recommended Plane for the default base plane of the legend.

        Args:
            max_speed: A number for the maximum wind speed along the axis
                in [m/s]. (Default: 5)
            direction: An optional number between 0 and 360 that represents the
                cardinal direction that the axis is facing in the XY
                plane. 0 = North, 90 = East, 180 = South, 270 = West. If None,
                the axis will simply be placed in the XY plane. (Default: None).
            base_point: A ladybug-geometry Point3D that represents the origin
                of the axis. (Default, (0, 0, 0)).
            length_dimension: A number to denote the length dimension of a 1 m/s
                wind vector in meters. (Default: 5).
            scale_factor: An optional number that will be multiplied by all dimensions
                to account for the fact that the wind profile may be displaying in
                a units system other than meters. (Default: 1).

        Returns:
            A ladybug-geometry Plane for the recommended legend plane.
        """
        direction = self._flip_direction(direction)
        bp, max_speed = base_point, int(max_speed)
        end_x = bp.x + ((max_speed + 2) * length_dimension * scale_factor)
        origin_pt = Point3D(end_x, bp.y, bp.z)
        if direction is None:
            return Plane(n=Vector3D(0, 0, 1), o=origin_pt, x=Vector3D(1, 0, 0))
        else:
            st_pl = Plane(n=Vector3D(0, -1, 0), o=origin_pt, x=Vector3D(1, 0, 0))
            dir_radians = math.radians(90) - math.radians(direction)
            return st_pl.rotate_xy(dir_radians, bp)

    def title_plane(
            self, direction=None, base_point=Point3D(0, 0, 0),
            length_dimension=5, scale_factor=1, text_height=None):
        """Get a recommended Plane for the default base plane of the title.

        Args:
            max_speed: A number for the maximum wind speed along the axis
                in [m/s]. (Default: 5)
            direction: An optional number between 0 and 360 that represents the
                cardinal direction that the axis is facing in the XY
                plane. 0 = North, 90 = East, 180 = South, 270 = West. If None,
                the axis will simply be placed in the XY plane. (Default: None).
            base_point: A ladybug-geometry Point3D that represents the origin
                of the axis. (Default, (0, 0, 0)).
            length_dimension: A number to denote the length dimension of a 1 m/s
                wind vector in meters. (Default: 5).
            scale_factor: An optional number that will be multiplied by all dimensions
                to account for the fact that the wind profile may be displaying in
                a units system other than meters. (Default: 1).
            text_height: An optional number for the height of the text in the axis
                label. If None, the text height will be inferred based on the
                length_dimension. (Default: None).

        Returns:
            A ladybug-geometry Plane for the recommended title plane.
        """
        direction = self._flip_direction(direction)
        bp = base_point
        step_d = length_dimension * scale_factor
        tick_d = step_d / 6
        txt_height = length_dimension / 2 if text_height is None else text_height
        sub_d = (tick_d * 2) + (txt_height * 5)
        if direction is None:
            origin_pt = Point3D(bp.x, bp.y - sub_d, bp.z)
            return Plane(n=Vector3D(0, 0, 1), o=origin_pt, x=Vector3D(1, 0, 0))
        else:
            origin_pt = Point3D(bp.x, bp.y, bp.z - sub_d)
            st_pl = Plane(n=Vector3D(0, -1, 0), o=origin_pt, x=Vector3D(1, 0, 0))
            dir_radians = math.radians(90) - math.radians(direction)
            return st_pl.rotate_xy(dir_radians, bp)

    def _check_terrain(self, value):
        """Check any string input for the terrain."""
        clean_input = value.lower()
        for key in self.TERRAINS:
            if key.lower() == clean_input:
                value = key
                break
        else:
            raise ValueError(
                'terrain {} is not recognized.\nChoose from the '
                'following:\n{}'.format(value, self.TERRAINS))
        return value

    def _compute_met_power_denom(self):
        """Compute the denominator of the power function."""
        h_ratio = self._met_boundary_layer_height / self._meteorological_height
        self._met_power_denom = h_ratio ** self._met_power_law_exponent

    def _compute_met_log_denom(self):
        """Compute the denominator of the logarithmic function."""
        h_ratio = self._meteorological_height / self._met_roughness_length
        self._met_log_denom = math.log(h_ratio)

    @staticmethod
    def _flip_direction(direction):
        """Flip the direction of a wind so it notes the orientation of arrows."""
        if direction is not None:
            if direction < 180:
                return direction + 180
            else:
                return direction - 180

    @staticmethod
    def _check_profile_inputs(max_height, vector_spacing):
        assert vector_spacing > 0, 'WindProfile vector spacing must be greater ' \
            'than zero. Got {}.'.format(vector_spacing)
        assert max_height > vector_spacing, 'WindProfile max height [{}] must be ' \
            'greater than vector spacing [{}].'.format(max_height, vector_spacing)

    @staticmethod
    def _frange(start, stop, step):
        """Range function capable of yielding float values."""
        while start < stop:
            yield start
            start += step

    def ToString(self):
        """Overwrite .NET ToString method."""
        return self.__repr__()

    def __repr__(self):
        """WindProfile representation."""
        return "WindProfile (terrain: {})".format(self.terrain)



"""A collection of helper functions used throughout the CLI.

Most functions assist with the serialization of objects to/from JSON.
"""
import json

from ladybug.analysisperiod import AnalysisPeriod


def _load_analysis_period_str(analysis_period_str):
    """Load an AnalysisPeriod from a JSON file.
    
    Args:
        analysis_period_str: A JSON file of an AnalysisPeriod to be loaded.
    """
    if analysis_period_str is not None and analysis_period_str != '' \
            and analysis_period_str != 'None':
        return AnalysisPeriod.from_string(analysis_period_str)



# coding=utf-8
"""Specific Heat Capacity data type."""
from __future__ import division

from .base import DataTypeBase


class SpecificHeatCapacity(DataTypeBase):
    """Specific Heat Capacity"""
    _units = ('J/kg-K', 'Btu/lb-F', 'kWh/kg-K', 'kBtu/lb-F', 'kJ/kg-K')
    _si_units = ('J/kg-K', 'kWh/kg-K', 'kJ/kg-K')
    _ip_units = ('Btu/lb-F', 'kBtu/lb-F')
    _min = 0
    _abbreviation = 'c'

    def _J_kgK_to_Btu_lbF(self, value):
        return value / 4186.8

    def _J_kgK_to_kWh_kgK(self, value):
        return value / 3600000.

    def _J_kgK_to_kBtu_lbF(self, value):
        return value / 4186800.

    def _J_kgK_to_kJ_kgK(self, value):
        return value / 1000.

    def _Btu_lbF_to_J_kgK(self, value):
        return value * 4186.8

    def _kWh_kgK_to_J_kgK(self, value):
        return value * 3600000.

    def _kBtu_lbF_to_J_kgK(self, value):
        return value * 4186800.

    def _kJ_kgK_to_J_kgK(self, value):
        return value * 1000.

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('J/kg-K', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        elif from_unit == 'kWh/kg-K':
            return self.to_unit(values, 'kBtu/lb-F', from_unit), 'kBtu/lb-F'
        else:
            return self.to_unit(values, 'Btu/lb-F', from_unit), 'Btu/lb-F'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        elif from_unit == 'kBtu/lb-F':
            return self.to_unit(values, 'kWh/kg-K', from_unit), 'kWh/kg-K'
        else:
            return self.to_unit(values, 'J/kg-K', from_unit), 'J/kg-K'



# coding=utf-8
"""Immutable versions of the Ladybug Data Collections.

Note that all of the methods or properties on an immutable collection that
return another data collection will return a collection that is mutable.

The only exceptions to this rule are:

*   duplicate() - which will always return an exact copy of the collection including
    its mutability.

*   get_aligned_collection() - which follows the mutability of the starting collection
    by default but includes an parameter to override this.

*   to_immutable() - which clearly always returns an immutable version of the collection

Note that the to_mutable() method on the immutable collections can always be used to
get a mutable version of an immutable collection.
"""
from __future__ import division

from .datacollection import HourlyDiscontinuousCollection, HourlyContinuousCollection, \
    DailyCollection, MonthlyCollection, MonthlyPerHourCollection


class _ImmutableCollectionBase(object):
    """Base class for all immutable Data Collections."""
    _mutable = False

    @property
    def values(self):
        """The Data Collection's list of numerical values."""
        return self._values

    @values.setter
    def values(self, values):
        if hasattr(self, '_values'):
            raise AttributeError(self._mutable_message)
        self._check_values(values)
        self._values = tuple(values)

    @property
    def _mutable_message(self):
        return 'values are immutable for {}.\nUse to_mutable() method to get a ' \
            'mutable version of this collection.'.format(self.__class__.__name__)

    def to_immutable(self):
        """Get an immutable version of this collection."""
        return self.duplicate()

    def duplicate(self):
        """Get a copy of this Data Collection."""
        collection = self.__class__(
            self.header.duplicate(), self._values, self.datetimes)
        collection._validated_a_period = self._validated_a_period
        return collection

    def __setitem__(self, key, value):
        raise AttributeError(self._mutable_message)


class HourlyDiscontinuousCollectionImmutable(
        _ImmutableCollectionBase, HourlyDiscontinuousCollection):
    """Immutable Discontinuous Data Collection at hourly or sub-hourly intervals."""

    def convert_to_culled_timestep(self, timestep=1):
        """This method is not available for immutable collections."""
        raise AttributeError(self._mutable_message)

    def to_mutable(self):
        """Get a mutable version of this collection."""
        new_obj = HourlyDiscontinuousCollection(self.header, self.values, self.datetimes)
        new_obj._validated_a_period = self._validated_a_period
        return new_obj


class HourlyContinuousCollectionImmutable(
        _ImmutableCollectionBase, HourlyContinuousCollection):
    """Immutable Continuous Data Collection at hourly or sub-hourly intervals."""

    def convert_to_culled_timestep(self, timestep=1):
        """This method is not available for immutable collections."""
        raise AttributeError(self._mutable_message)

    def to_mutable(self):
        """Get a mutable version of this collection."""
        new_obj = HourlyContinuousCollection(self.header, self.values)
        new_obj._validated_a_period = self._validated_a_period
        return new_obj

    def duplicate(self):
        """Return a copy of the current Data Collection."""
        return self.__class__(self.header.duplicate(), self._values)


class DailyCollectionImmutable(
        _ImmutableCollectionBase, DailyCollection):
    """Immutable Daily Data Collection."""

    def to_mutable(self):
        """Get a mutable version of this collection."""
        new_obj = DailyCollection(self.header, self.values, self.datetimes)
        new_obj._validated_a_period = self._validated_a_period
        return new_obj


class MonthlyCollectionImmutable(
        _ImmutableCollectionBase, MonthlyCollection):
    """Immutable Monthly Data Collection."""

    def to_mutable(self):
        """Get a mutable version of this collection."""
        new_obj = MonthlyCollection(self.header, self.values, self.datetimes)
        new_obj._validated_a_period = self._validated_a_period
        return new_obj


class MonthlyPerHourCollectionImmutable(
        _ImmutableCollectionBase, MonthlyPerHourCollection):
    """Immutable Monthly Per Hour Data Collection."""

    def to_mutable(self):
        """Get a mutable version of this collection."""
        new_obj = MonthlyPerHourCollection(self.header, self.values, self.datetimes)
        new_obj._validated_a_period = self._validated_a_period
        return new_obj



# coding=utf-8
"""Density data type."""
from __future__ import division

from .base import DataTypeBase


class Density(DataTypeBase):
    """Density
    """
    _units = ('kg/m3', 'lb/ft3', 'g/cm3', 'oz/in3')
    _si_units = ('kg/m3', 'g/cm3')
    _ip_units = ('lb/ft3', 'oz/in3')
    _min = 0
    _abbreviation = 'rho'

    def _kg_m3_to_lb_ft3(self, value):
        return value * 0.062428

    def _kg_m3_to_g_cm3(self, value):
        return value * 0.001

    def _kg_m3_to_oz_in3(self, value):
        return value * 0.000578037

    def _lb_ft3_to_kg_m3(self, value):
        return value / 0.062428

    def _g_cm3_to_kg_m3(self, value):
        return value / 0.001

    def _oz_in3_to_kg_m3(self, value):
        return value / 0.000578037

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('kg/m3', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        elif from_unit == 'g/cm3':
            return self.to_unit(values, 'oz/in3', from_unit), 'oz/in3'
        else:
            return self.to_unit(values, 'lb/ft3', from_unit), 'lb/ft3'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        elif from_unit == 'oz/in3':
            return self.to_unit(values, 'g/cm3', from_unit), 'g/cm3'
        else:
            return self.to_unit(values, 'kg/m3', from_unit), 'kg/m3'



# coding=utf-8
"""Conductance data type."""
from __future__ import division

from .base import DataTypeBase


class Conductance(DataTypeBase):
    """Conductance
    """
    _units = ('W/K', 'Btu/h-F')
    _si_units = ('W/K')
    _ip_units = ('Btu/h-F')
    _min = 0
    _abbreviation = 'G'

    def _W_K_to_Btu_hF(self, value):
        return value / 0.5271751322

    def _Btu_hF_to_W_K(self, value):
        return value * 0.5271751322

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('W/K', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit == 'Btu/h-F':
            return values, from_unit
        else:
            return self.to_unit(values, 'Btu/h-F', from_unit), 'Btu/h-F'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit == 'W/K':
            return values, from_unit
        else:
            return self.to_unit(values, 'W/K', from_unit), 'W/K'



# coding=utf-8
"""Resistance data type."""
from __future__ import division

from .base import DataTypeBase


class Resistance(DataTypeBase):
    """Resistance
    """
    _units = ('K/W', 'F-h/Btu')
    _si_units = ('K/W',)
    _ip_units = ('F-h/Btu',)
    _min = 0
    _abbreviation = 'R'

    def _K_W_to_Fh_Btu(self, value):
        return value * 0.5271751322

    def _Fh_Btu_to_K_W(self, value):
        return value / 0.5271751322

    def to_unit(self, values, unit, from_unit):
        """Return values converted to the unit given the input from_unit."""
        return self._to_unit_base('K/W', values, unit, from_unit)

    def to_ip(self, values, from_unit):
        """Return values in IP and the units to which the values have been converted."""
        if from_unit in self.ip_units:
            return values, from_unit
        else:
            return self.to_unit(values, 'F-h/Btu', from_unit), 'F-h/Btu'

    def to_si(self, values, from_unit):
        """Return values in SI and the units to which the values have been converted."""
        if from_unit in self.si_units:
            return values, from_unit
        else:
            return self.to_unit(values, 'K/W', from_unit), 'K/W'
